"""
Newspaper and magazine articles. By default considered uninteresting.
"""
from pathlib import Path

from epstein_files.documents.config.categories.legal import JANE_DOE_V_EPSTEIN_TRUMP
from epstein_files.documents.config.config_builder import JANE_DOE_V_USA, wolff_draft_cfg
from epstein_files.documents.config.doc_cfg import DocCfg
from epstein_files.people.names import *
from epstein_files.util.constant.strings import *
from epstein_files.util.helpers.string_helper import join_truthy, quote

SHIMON_POST_ARTICLE = f'selection of articles about the mideast'
VALAR_FUND = f"{PETER_THIEL}'s {VALAR_VENTURES} fund"


ARTICLE_CFGS = [
    DocCfg(id='030013', author='Aviation International News', description=f'article', date='2012-07-01'),
    DocCfg(id='013275', author=BLOOMBERG, description=f"article on notable 2013 obituaries", date='2013-12-26'),
    DocCfg(id='026543', author=BLOOMBERG, description=f"BNA article about taxes"),
    DocCfg(id='014865', author='Boston Globe', description=f"article about {ALAN_DERSHOWITZ}"),
    DocCfg(id='033231', author='Business Standard', description=f"article about Trump's visit with India's Modi"),
    DocCfg(id='023572', author=CHINA_DAILY, description=f"article on China's Belt & Road Initiative"),
    DocCfg(id='023571', author=CHINA_DAILY, description=f"article on terrorism, Macau, trade initiatives", date='2016-09-18'),
    DocCfg(id='023570', author=CHINA_DAILY, description=f"article on Belt & Road in Central/South America, Xi philosophy", date='2017-05-14'),
    DocCfg(id='025115', author=CHINA_DAILY, description=f"article on China and the US working together", date='2017-05-14'),
    DocCfg(id='012704', author='Daily Business Review', description=f"article about {JANE_DOE_V_USA} and {CVRA}", date='2011-04-21'),
    DocCfg(id='025292', author=DAILY_MAIL, description=f"article on Bill Clinton being named in a lawsuit"),
    DocCfg(id='019468', author=DAILY_MAIL, description=f"article on Epstein and Clinton"),
    DocCfg(id='022970', author=DAILY_MAIL, description=f"article on Epstein and Prince Andrew"),
    DocCfg(id='031186', author=DAILY_MAIL, description=f'article on allegations of rape of 13 year old against Trump', date='2016-11-02'),
    DocCfg(id='013437', author=DAILY_TELEGRAPH, description=f"article about Epstein's diary", date='2011-03-05'),
    DocCfg(id='023287', author=DAILY_TELEGRAPH, description=f"article about a play based on the Oslo Accords", date='2017-09-15'),
    DocCfg(id='019206', author=EDWARD_JAY_EPSTEIN, description=f"WSJ article about Edward Snowden", date='2016-12-30'),
    DocCfg(id='023567', author='Financial Times', description=f"article about quantitative easing"),
    DocCfg(id='026761', author='Forbes', description=f"article about {BARBRO_C_EHNBOM} 'Swedish American Group Focuses On Cancer'"),
    DocCfg(id='031716', author='Fortune Magazine', description=f'article by {TOM_BARRACK}', date='2016-10-22'),
    DocCfg(id='019444', author='Frontlines magazine', description=f"article 'Biologists Dig Deeper'", date='2008-01-01'),
    DocCfg(id='023720', author='Future Science', description=f'article: "Is Shame Necessary?" by {JENNIFER_JACQUET}'),
    DocCfg(id='021094', author='Globe and Mail', description=f"article about Gerd Heinrich"),
    DocCfg(id='013268', author='JetGala', description=f"article about airplane interior designer {ERIC_ROTH}"),
    DocCfg(id='029539', author=LA_TIMES, description=f"Alan Trounson interview on California stem cell research and CIRM"),
    DocCfg(id='029865', author=LA_TIMES, description=f"front page article about {DEEPAK_CHOPRA} and young Iranians", date='2016-11-05'),
    DocCfg(id='026598', author=LA_TIMES, description=f"op-ed about why America needs a Ministry of Culture"),
    DocCfg(id='027024', author=LA_TIMES, description=f"Scientists Create Human Embryos to Make Stem Cells", date='2013-05-15'),
    DocCfg(id='022811', author='Law.com', description='Sarah Ransome Identifies Herself in Epstein Sex Trafficking Case', date='2018-01-09'),
    DocCfg(id='031776', author='Law360', description=f"article about Michael Avenatti by Andrew Strickler"),
    DocCfg(id='023102', author='Litigation Daily', description=f"article about {REID_WEINGARTEN}", date='2015-09-04'),
    DocCfg(id='029340', author='MarketWatch', description=f'article about estate taxes, particularly Epstein\'s favoured GRATs'),
    # wolff_draft_cfg
    wolff_draft_cfg('022707'),
    wolff_draft_cfg('022727'),
    wolff_draft_cfg('022746'),
    wolff_draft_cfg('022844'),
    wolff_draft_cfg('022863'),
    wolff_draft_cfg('022952'),
    wolff_draft_cfg('024229'),
    wolff_draft_cfg(
        '022894',
        f"claiming Kelimbetov, {BROCK_PIERCE}, and {LARRY_SUMMERS} all visited Epstein around the time Tether was founded",
        show_with_name=BROCK_PIERCE,
        truncate_to=(13_500, 16_000),
    ),
    DocCfg(id='031198', author='Morning News USA', description=f"article about identify of Jane Doe in {JANE_DOE_V_EPSTEIN_TRUMP}"),
    DocCfg(id='015462', author='Nautilus Education', description=f'magazine (?) issue'),
    DocCfg(id='031972', author=NYT, description=f"article about #MeToo allegations against {LAWRENCE_KRAUSS}", date='2018-03-07'),
    DocCfg(id='032435', author=NYT, description=f'article about Chinese butlers'),
    DocCfg(id='029452', author=NYT, description=f"article about {PETER_THIEL}"),
    DocCfg(id='025328', author=NYT, description=f"article about radio host Bob Fass and Robert Durst"),
    DocCfg(id='033479', author=NYT, description=f"article about Rex Tillerson", date='2010-03-14'),
    DocCfg(id='028481', author=NYT, description=f'article about {STEVE_BANNON}', date='2018-03-09'),
    DocCfg(id='033181', author=NYT, description=f"article about Trump's tax avoidance", date='2016-10-31'),
    DocCfg(id='023097', author=NYT, description=f"column about The Aristocrats by Frank Rich 'The Greatest Dirty Joke Ever Told'"),
    DocCfg(id='033365', author=NYT, description=f'column about trade war with China by Kevin Rudd'),
    DocCfg(id='019439', author=NYT, description=f"column about the Clintons and money by Maureen Dowd", date='2013-08-17'),
    DocCfg(id='EFTA00040076', author=NYT, description=f"article about Epstein's death", date='2019-07-08'),
    DocCfg(id='029925', author='New Yorker', description=f"article about the placebo effect by Michael Specter"),
    DocCfg(id='013435', author=PALM_BEACH_DAILY_NEWS, description=f"article about Epstein's address book", date='2011-03-11'),
    DocCfg(id='013440', author=PALM_BEACH_DAILY_NEWS, description=f"article about Epstein's gag order", date='2011-07-13'),
    DocCfg(id='029238', author=PALM_BEACH_DAILY_NEWS, description=f"article about Epstein's plea deal"),
    DocCfg(id='021775', author=PALM_BEACH_POST, description="article about 'He Was 50. And They Were Girls'", attached_to_email_id='021764'),
    DocCfg(id='022989', author=PALM_BEACH_POST, description="article about alleged rape of 13 year old by Trump"),
    DocCfg(id='022987', author=PALM_BEACH_POST, description="article about just a headline on Trump and Epstein"),
    DocCfg(id='015028', author=PALM_BEACH_POST, description="article about reopening Epstein's criminal case"),
    DocCfg(id='022990', author=PALM_BEACH_POST, description="article about Trump and Epstein"),
    DocCfg(id='031753', author=PAUL_KRASSNER, description=f'essay for Playboy in the 1980s', date='1985-01-01'),
    DocCfg(id='023638', author=PAUL_KRASSNER, description=f'magazine interview'),
    DocCfg(id='024374', author=PAUL_KRASSNER, description=f"Remembering Cavalier Magazine"),
    DocCfg(id='030187', author=PAUL_KRASSNER, description=f'"Remembering Lenny Bruce While We\'re Thinking About Trump" (draft?)'),
    DocCfg(id='019088', author=PAUL_KRASSNER, description=f'"Are Rape Jokes Funny?" (draft)', date='2012-07-28'),
    DocCfg(id='012740', author=PEGGY_SIEGAL, description=f"article about Venice Film Festival"),
    DocCfg(id='013442', author=PEGGY_SIEGAL, description=f"draft about Oscars", date='2011-02-27'),
    DocCfg(id='012700', author=PEGGY_SIEGAL, description=f"film events diary", date='2011-02-27'),
    DocCfg(id='012690', author=PEGGY_SIEGAL, description=f"film events diary early draft of 012700", date='2011-02-27'),
    DocCfg(id='013450', author=PEGGY_SIEGAL, description=f"Oscar Diary in Avenue Magazine", date='2011-02-27'),
    DocCfg(id='010715', author=PEGGY_SIEGAL, description=f"Oscar Diary April", date='2012-02-27'),
    DocCfg(id='019849', author=PEGGY_SIEGAL, description=f"Oscar Diary April", date='2017-02-27', duplicate_ids=['019864']),
    DocCfg(id='026851', author='Politifact', description=f"lying politicians chart", date='2016-07-26'),
    DocCfg(id='033253', author=ROBERT_LAWRENCE_KUHN, description=f'{BBC} article about Rohingya in Myanmar', attached_to_email_id='033252'),
    DocCfg(id='026887', author=ROBERT_LAWRENCE_KUHN, description=f'{BBC} "New Tariffs - Trade War"'),
    DocCfg(id='026877', author=ROBERT_LAWRENCE_KUHN, description=f'{CNN} "New Tariffs - Trade War"'),
    DocCfg(id='026868', author=ROBERT_LAWRENCE_KUHN, description=f'{CNN} "Quest Means Business New China Tariffs — Trade War"', date='2018-09-18'),
    DocCfg(id='023707', author=ROBERT_LAWRENCE_KUHN, description=f'{CNN} "Quest Means Business U.S. and China Agree to Pause Trade War"', date='2018-12-03'),
    DocCfg(id='029176', author=ROBERT_LAWRENCE_KUHN, description=f'{CNN} "U.S. China Tariffs - Trade War"', attached_to_email_id='029174'),
    DocCfg(id='032638', author=ROBERT_LAWRENCE_KUHN, description=f'{CNN} "Xi Jinping and the New Politburo Committee"', attached_to_email_id='032637'),
    DocCfg(id='023666', author=ROBERT_LAWRENCE_KUHN, description=f"sizzle reel / television appearances", date='2018-09-30', attached_to_email_id='033252'),
    DocCfg(id='016996', author='SciencExpress', description=f'article "Quantitative Analysis of Culture Using Millions of Digitized Books" by Jean-Baptiste Michel'),
    DocCfg(id='025104', author='SCMP', description=f"article about China and globalisation"),
    DocCfg(id='030030', author=SHIMON_POST, description=SHIMON_POST_ARTICLE, date='2011-03-29'),
    DocCfg(id='025610', author=SHIMON_POST, description=SHIMON_POST_ARTICLE, date='2011-04-03'),
    DocCfg(id='023458', author=SHIMON_POST, description=SHIMON_POST_ARTICLE, date='2011-04-17'),
    DocCfg(id='023487', author=SHIMON_POST, description=SHIMON_POST_ARTICLE, date='2011-04-18'),
    DocCfg(id='030531', author=SHIMON_POST, description=SHIMON_POST_ARTICLE, date='2011-05-16'),
    DocCfg(id='024958', author=SHIMON_POST, description=SHIMON_POST_ARTICLE, date='2011-05-08'),
    DocCfg(id='030060', author=SHIMON_POST, description=SHIMON_POST_ARTICLE, date='2011-05-13'),
    DocCfg(id='031834', author=SHIMON_POST, description=SHIMON_POST_ARTICLE, date='2011-05-16'),
    DocCfg(id='023517', author=SHIMON_POST, description=SHIMON_POST_ARTICLE, date='2011-05-26'),
    DocCfg(id='030268', author=SHIMON_POST, description=SHIMON_POST_ARTICLE, date='2011-05-29'),
    DocCfg(id='029628', author=SHIMON_POST, description=SHIMON_POST_ARTICLE, date='2011-06-04'),
    DocCfg(id='018085', author=SHIMON_POST, description=SHIMON_POST_ARTICLE, date='2011-06-07'),
    DocCfg(id='030156', author=SHIMON_POST, description=SHIMON_POST_ARTICLE, date='2011-06-22'),
    DocCfg(id='031876', author=SHIMON_POST, description=SHIMON_POST_ARTICLE, date='2011-06-14'),
    DocCfg(id='032171', author=SHIMON_POST, description=SHIMON_POST_ARTICLE, date='2011-06-26'),
    DocCfg(id='029932', author=SHIMON_POST, description=SHIMON_POST_ARTICLE, date='2011-07-03'),
    DocCfg(id='031913', author=SHIMON_POST, description=SHIMON_POST_ARTICLE, date='2011-08-24'),
    DocCfg(id='024592', author=SHIMON_POST, description=SHIMON_POST_ARTICLE, date='2011-08-25'),
    DocCfg(id='024997', author=SHIMON_POST, description=SHIMON_POST_ARTICLE, date='2011-09-08'),
    DocCfg(id='031941', author=SHIMON_POST, description=SHIMON_POST_ARTICLE, date='2011-11-17'),
    DocCfg(id='030829', author='South Florida Sun Sentinel', description=f'article about {BRAD_EDWARDS} and {JEFFREY_EPSTEIN}'),
    DocCfg(id='021092', author='Tatler', description=f'single page of article about {GHISLAINE_MAXWELL} shredding documents'),
    DocCfg(id='030333', author='The Independent', description=f'article about Prince Andrew, Epstein, and Epstein\'s butler who stole his address book'),
    DocCfg(id='010754', author='U.S. News', description=f"article about Yitzhak Rabin"),
    DocCfg(id='014498', author=VI_DAILY_NEWS, description='article', date='2016-12-13'),
    DocCfg(id='031171', author=VI_DAILY_NEWS, description='article', date='2019-02-06'),
    DocCfg(id='023048', author=VI_DAILY_NEWS, description='article', date='2019-02-27'),
    DocCfg(id='023046', author=VI_DAILY_NEWS, description='article', date='2019-02-27'),
    DocCfg(id='031170', author=VI_DAILY_NEWS, description='article', date='2019-03-06'),
    DocCfg(id='016506', author=VI_DAILY_NEWS, description='article', date='2019-02-28'),
    DocCfg(id='018862', author=VI_DAILY_NEWS, description='articles about Sen. Alvin Williams Jr. Fraud case, arson', date='2012-11-09'),
    DocCfg(id='016507', author=VI_DAILY_NEWS, description=f'"Perversion of Justice" by {JULIE_K_BROWN}', date='2018-12-19'),
    DocCfg(id='019212', author=WAPO, description=f'and Times Tribune articles about Bannon, Trump, and healthcare execs'),
    DocCfg(id='033379', author=WAPO, description=f'"How Washington Pivoted From Finger-Wagging to Appeasement" (about Viktor Orban)', date='2018-05-25'),
    DocCfg(
        id='031396',
        author=WAPO,
        description=f"DOJ discipline office with limited reach to probe handling of controversial sex abuse case",
        date='2019-02-06',
        duplicate_ids=['031415'],
    ),
    DocCfg(id='030199', description=f'article about Trump rape allegations in {JANE_DOE_V_EPSTEIN_TRUMP}', date='2017-11-16'),
    DocCfg(id='031725', description=f"article about Gloria Allred and Trump allegations", date='2016-10-10'),
    DocCfg(id='026648', description=f'article about {JASTA} lawsuit against Saudi Arabia by 9/11 victims (Russian propaganda?)', date='2017-05-13'),
    DocCfg(id='032159', description=f"article about microfinance and cell phones in Zimbabwe, Strive Masiyiwa (Econet Wireless)"),
    DocCfg(id='030825', description=f'{ARTICLE_DRAFT} Syria'),
    DocCfg(id='027051', description=f"German article about the 2013 Lifeball / AIDS Gala", date='2013-01-01', attached_to_email_id='027049'),
    DocCfg(id='033480', description=f"John Bolton press clipping", date='2018-04-06', duplicate_ids=['033481']),
    DocCfg(id='013403', description=f"{LEXIS_NEXIS} result from The Evening Standard about Bernie Madoff", date='2009-12-24'),
    DocCfg(id='021093', description=f"page of unknown article about Epstein and Maxwell"),
    DocCfg(id='031191', description=f"single page of unknown article about Epstein and Trump's relationship in 1997"),
    DocCfg(id='026520', description=f'Spanish article about {SULTAN_BIN_SULAYEM}', date='2013-09-27'),
    DocCfg(
        id='031736',
        author='Abdulnaser Salamah',
        description=f'{TRANSLATION} Arabic article "Trump; Prince of Believers (Caliph)!"',
        date='2017-05-13',
    ),
    DocCfg(id='025094', description=f'{TRANSLATION} Spanish article about Cuba', date='2015-11-08'),
    DocCfg(id='031794', description=f"very short French magazine clipping"),

    # DOJ files
    DocCfg(
        id='EFTA02711825',
        author='Techonomy',
        author_uncertain='either this or EFTA01139627 are referenced in email EFTA01745739',
        description='article titled "Artificial Intelligence Ignites in Ethiopia"',
        is_interesting=True,
    ),
    DocCfg(
        id='EFTA01139627',
        author='Techonomy',
        author_uncertain='either this or EFTA02711825 are referenced in email EFTA01745739',
        date='2014-08-15',
        date_uncertain=f'using date of {BEN_GOERTZEL} email',
        description='article about iCog Labs titled "A Harvard Financier, Jeffrey Epstein, Advances Artificial Intelligence in Ethiopia"',
        is_interesting=True,
    ),
]



import re
from typing import Literal

# Types
PartialName = Literal['both', 'first', 'last']

# Categories
ACADEMIA = 'academia'
ARTS = 'arts'
BUSINESS = 'business'
FRIEND = 'friend'
JOURNALIST = 'journalist'
LAWYER = 'lawyer'
LOBBYIST = 'lobbyist'
MIDEAST = 'mideast'
PUBLICIST = 'publicist'
REAL_ESTATE = 'real estate'
REPUTATION = 'reputation'
REPUTATION_MGMT = f'{REPUTATION} management'
TECH_BRO = 'tech'

# Locations
PALM_BEACH = 'Palm Beach'
VIRGIN_ISLANDS = 'Virgin Islands'

# Highlighter
HIGHLIGHTED_QUOTE = 'highlighted_quote'
REGEX_STYLE_PREFIX = 'regex'

# Styles
ALT_LINK_STYLE = 'white dim'
ARCHIVE_LINK_COLOR = 'slate_blue3'
ARCHIVE_LINK_UNDERLINE = f"{ARCHIVE_LINK_COLOR} underline"
ARCHIVE_ALT_LINK_STYLE = 'medium_purple4 italic'  # "(raw files)" link
AUX_SITE_LINK_STYLE = 'dark_orange3'
DEFAULT_NAME_STYLE = 'grey23'
EPSTEIN_COLOR = 'blue1'
EXCERPT_STYLE = 'italic cornsilk1'
SOCIAL_MEDIA_LINK_STYLE = 'pale_turquoise4'
SUBHEADER_STYLE = 'dim gray54 italic'
TEXT_LINK = 'text_link'
TIMESTAMP_STYLE = 'turquoise4'
TIMESTAMP_DIM = f"turquoise4 dim"
VICTIM_LAWYER = 'victim lawyer'

# File IDs
LEON_BLACK_EMAIL_ID = '023208'

# Misc
APPEARS_IN = 'appears in'
ARTICLE_DRAFT = 'draft of an article about'
AUTHOR = 'author'
CHRONOLOGICAL = 'chronological'
DEFAULT = 'default'
EMAIL = 'email'
JEE = 'JEE'
JSON = 'json'
LAW_ENFORCEMENT = 'law enforcement'
NA = 'n/a'
REDACTED = '<REDACTED>'
QUESTION_MARKS = '(???)'
SCREENSHOT = 'screenshot of'
TEXT_MESSAGE = 'text message'
TRANSLATION = 'translation of'
VICTIM_DIARY = 'victim diary'

# CLI args
SUPPRESS_OUTPUT = '--suppress-output'

# Document subclass names (this sucks)
DOCUMENT_CLASS = 'Document'
DOJ_FILE_CLASS = 'DojFile'
EMAIL_CLASS = 'Email'
JSON_FILE_CLASS = 'JsonFile'
MESSENGER_LOG_CLASS = 'MessengerLog'
OTHER_FILE_CLASS = 'OtherFile'

# Law
CVRA = "Crime Victims' Rights Act [CVRA]"
JASTA = 'JASTA'
JASTA_FULL = f"{JASTA} (Justice Against Sponsors of Terrorism Act)"

# Organizations
BROCKMAN_INC = 'Brockman, Inc.'
HARVARD = 'Harvard'
HARVARD_POETRY = f'{HARVARD} poetry stuff from Lisa New'  # TODO: name be here?
PALM_BEACH_CODE_ENFORCEMENT = f'{PALM_BEACH} Code Enforcement'
PALM_BEACH_PROPERTY_INFO = f"{PALM_BEACH} property info"
PALM_BEACH_TSV = f"TSV of {PALM_BEACH} property"
PALM_BEACH_WATER_COMMITTEE = f'{PALM_BEACH} Water Committee'
UN_GENERAL_ASSEMBLY = '67th U.N. General Assembly'

# Tranches
DOJ_2026_TRANCHE = '2026 DOJ tranche'
HOUSE_OVERSIGHT_TRANCHE = '2025 Oversight tranche'
EPSTEIN_FILES_NOV_2025 = 'epstein_files_nov_2025'

# Regexes / file names
EFTA_PREFIX = 'EFTA'
HOUSE_OVERSIGHT_PREFIX = 'HOUSE_OVERSIGHT_'
DOJ_DATASET_ID_REGEX = re.compile(r"(?:epstein_dataset_|DataSet )(\d+)")
DOJ_FILE_STEM_REGEX = re.compile(fr"({EFTA_PREFIX}\d{{8}}(_\d{{1,3}})?)")
DOJ_FILE_NAME_REGEX = re.compile(fr"{DOJ_FILE_STEM_REGEX.pattern}(\.txt)?")
HOUSE_OVERSIGHT_NOV_2025_ID_REGEX = re.compile(r"(\d{6}(_\d{1,3})?)")
HOUSE_OVERSIGHT_NOV_2025_FILE_STEM_REGEX = re.compile(fr"{HOUSE_OVERSIGHT_PREFIX}{HOUSE_OVERSIGHT_NOV_2025_ID_REGEX.pattern}")
HOUSE_OVERSIGHT_NOV_2025_FILE_NAME_REGEX = re.compile(fr"{HOUSE_OVERSIGHT_NOV_2025_FILE_STEM_REGEX.pattern}(\.txt(\.json)?)?")
LOCAL_EXTRACT_REGEX = re.compile(r"_\d{1,2}$")
QUESTION_MARKS_REGEX = re.compile(fr' {re.escape(QUESTION_MARKS)}$')

# Other regexes
CASE_ID_REGEX = re.compile(r"Case\s+(Number:\s+)?\d:\d{2}-[a-z]{2}-\d{5}-[A-Z]{3}")

# Decorative
LEFT_ARROWS = '⇚ ⇠ ⇷ ⟵ ⇜ ⇽ ⬰ ⬲'
RIGHT_ARROWS = '↦ ↠ ➔ ➤ ➦ ➱ ➪ ➾'
LINES = '-﹊﹌﹉﹏﹎＿'

# Whitespace
INDENT = '    '
INDENT_NEWLINE = f'\n{INDENT}'
INDENTED_JOIN = f',{INDENT_NEWLINE}'

WEEKDAYS = 'Sunday Monday Tuesday Wednesday Thursday Friday Saturday'.split()
MONTHS = 'January February March April May June July August September October November December'.split()

STATES_PIPE_DELIMITED = """
AL|Alabama
AK|Alaska
AZ|Arizona
AR|Arkansas
CA|California
CO|Colorado
CT|Connecticut
DE|Delaware
FL|Florida
GA|Georgia
HI|Hawaii
ID|Idaho
IL|Illinois
IN|Indiana
IA|Iowa
KS|Kansas
KY|Kentucky
LA|Louisiana
ME|Maine
MD|Maryland
MA|Massachusetts
MI|Michigan
MN|Minnesota
MS|Mississippi
MO|Missouri
MT|Montana
NE|Nebraska
NY|New York
NV|Nevada
NH|New Hampshire
NJ|New Jersey
NM|New Mexico
NC|North Carolina
ND|North Dakota
OH|Ohio
OK|Oklahoma
OR|Oregon
PA|Pennsylvania
RI|Rhode Island
SC|South Carolina
SD|South Dakota
TN|Tennessee
TX|Texas
UT|Utah
VT|Vermont
VA|Virginia
WA|Washington
WV|West Virginia
WI|Wisconsin
WY|Wyoming
""".strip()

OTHER_US_DELIMITED = """
DC|District of Columbia
AS|American Samoa
GU|Guam
MP|Northern Mariana Islands
PR|Puerto Rico
UM|United States Minor Outlying Islands
VI|Virgin Islands, U.S.
""".strip()

STATE_CODES = {
    row.split('|')[0]: row.split('|')[1]
    for row in STATES_PIPE_DELIMITED.split('\n')
}

# some states excluded for pattern matching reasons re: important locations or names
STATE_NAME_PATTERNS = [
    f"{state}n?" if state.endswith('a') else state
    for state in STATE_CODES.values()
    if state not in ['Arizon', 'Florida', 'New York', 'Texas', 'Virginia', 'Washington']
]



from collections import defaultdict

from rich.padding import Padding
from rich.text import Text

from epstein_files.documents.emails.constants import FALLBACK_TIMESTAMP
from epstein_files.util.helpers.string_helper import quote
from epstein_files.util.logging import logger


def highlight_config_dict() -> dict:
    from epstein_files.output.highlight_config import HIGHLIGHTED_NAMES

    return {
        hn.label: {
            'patterns': hn.patterns,
            'regex.pattern': hn.regex.pattern,
        }
        for hn in HIGHLIGHTED_NAMES
    }


def print_all_highlighted_quotes(epstein_files):
    from epstein_files.output.rich import console

    for email in epstein_files.emails:
        if email.config and email.config.highlight_quote:
            console.print(email)


def print_all_timestamps(epstein_files):
    from epstein_files.output.rich import console
    fallbacks = valid = 0

    for i, doc in enumerate(epstein_files.unique_documents):
        console.print(doc._summary)

        if doc.timestamp == FALLBACK_TIMESTAMP:
            fallbacks += 1
        elif doc.timestamp:
            valid += 1

    no_timestamp = len(epstein_files.unique_documents) - valid - fallbacks
    console.print(f"\nFound {i + 1} documents (no_timestamp={no_timestamp}, valid={valid}, fallback={fallbacks})\n")


def print_highlight_patterns() -> None:
    from epstein_files.output.rich import print_json
    print_json('patterns', highlight_config_dict())
    import sys
    sys.exit()


def print_file_counts(epstein_files) -> None:
    from epstein_files.output.rich import console, styled_dict
    counts = defaultdict(int)

    for i, doc in enumerate(epstein_files.unique_documents):
        counts[doc._class_name] += 1

    counts['total'] = i + 1
    console.print(styled_dict(counts))


def _verify_filenames(epstein_files):
    doc_filenames = set([doc.file_path.name for doc in epstein_files.documents])

    for file_path in epstein_files.file_paths:
        if file_path.name not in doc_filenames:
            print(f"'{file_path}' is not in list of {len(doc_filenames)} Document obj filenames!")


def print_interesting_doc_panels_and_props(epstein_files, sort_by_category: bool = True):
    """Only prints OtherFile objects."""
    from epstein_files.documents.other_file import OtherFile
    from epstein_files.output.rich import console, indent_txt, styled_dict
    num_printed = 0
    num_interesting = 0

    if sort_by_category:
        docs = sorted(epstein_files.documents, key=lambda d: [d.category, d.timestamp or FALLBACK_TIMESTAMP])
    else:
        docs = epstein_files.documents

    for doc in docs:
        if not isinstance(doc, OtherFile):
            continue
        elif doc.is_interesting is None and doc.config is None:
            continue

        if doc.config:
            props = doc.config.truthy_props
            props.pop('id')

            if doc.is_interesting != doc.config.is_of_interest:
                logger.warning(f"mismatch of config.is_of_interest and doc.is_interesting")
                props['doc.is_interesting'] = doc.is_interesting
        else:
            props = {'doc.is_interesting': doc.is_interesting}

        # console.print(doc._summary_panel)
        console.print(doc.rich_header())
        from epstein_files.util.env import args

        if args.debug:
            props_table = styled_dict(props, sep=': ')
            props_table = Padding(indent_txt(props_table, 12), (0, 0, 1, 0))
            console.print(props_table)

        num_printed += 1
        num_interesting += int(doc.is_interesting or False)
        # console.line()

    print('')
    logger.warning(f"Printed {num_printed} object configs, {num_interesting} interesting ones.\n\n")



from dataclasses import dataclass, field
from pathlib import Path

from rich import box
from rich.align import Align
from rich.console import Group
from rich.padding import Padding
from rich.panel import Panel
from rich.text import Text

from epstein_files.documents.communication import Communication
from epstein_files.documents.document import Document
from epstein_files.documents.doj_file import DojFile
from epstein_files.documents.email import Email
from epstein_files.documents.messenger_log import MessengerLog
from epstein_files.documents.other_file import OtherFile
from epstein_files.output.layout_elements.file_display import BasePanel, FileDisplay
from epstein_files.output.html.builder import buffer_as_html, rich_to_html, table_to_html, write_templated_html
from epstein_files.output.html.elements import div_class, to_em
from epstein_files.output.rich import console
from epstein_files.people.person import PEOPLE_BIOS, Person
from epstein_files.people.names import *
from epstein_files.util.env import args, site_config
from epstein_files.util.helpers.data_helpers import uniq_sorted
from epstein_files.util.logging import logger
from epstein_files.util.timer import Timer

OTHER_FILES_TABLE_MSG = Text("(non emails will appear in tables)", 'gray27 italic')


@dataclass
class DocPrinter:
    """Handles printing collections of documents with biographical info panels interspersed."""
    collect_other_files_to_tables: bool = True
    html_elements: list[str] = field(default_factory=list)
    other_files_queue: list[OtherFile] = field(default_factory=list)
    last_bio_panel = ''
    people_encountered: set[str] = field(default_factory=set)
    printed_docs: list[Document] = field(default_factory=list)
    printed_file_displays: list[FileDisplay] = field(default_factory=list)
    suppressed_docs_queue: list[Document] = field(default_factory=list)

    @property
    def printed_emails(self) -> list[Email]:
        emails = [e for e in self.printed_docs if isinstance(e, Email)]

        if len(emails) != len(self.printed_docs):
            logger.warning(f"DocPrinted.printed_emails returning {len(emails)} of {len(self.printed_docs)} printed docs...")

        return emails

    def new_names(self, document: Document) -> list[str]:
        return [p for p in document.people if p in PEOPLE_BIOS and p not in self.people_encountered]

    def print_characters_panel(self, names: list[str], is_sticky: bool = True) -> None:
        if (bio_panel := self._biographical_panel(uniq_sorted(names))):
            if self.last_bio_panel:
                raise RuntimeError(f"last_bio_panel should be empty, instead it's:\n{self.last_bio_panel}")

            console.print(self._align_biographical_panel(bio_panel))
            self.people_encountered.update(names)
            class_name = ('sticky_' if is_sticky else '') + 'person_bio_panel'
            self.last_bio_panel = div_class(rich_to_html(bio_panel, minimize_width=True), class_name)

    def print_documents(self, docs: Sequence[Document | FileDisplay]) -> None:
        if len(self.html_elements) == 0:
            self.html_elements.append(self._html_so_far())

        if len(docs) > 500:
            logger.warning(f"{type(self).__name__}.print_documents() called with {len(docs):,} Documents...")

        timer = Timer()
        i = 0

        for i, doc in enumerate(docs, 1):
            logger.info(f"Printing {doc}")

            if isinstance(doc, Document) and doc.suppressed_txt:
                self.suppressed_docs_queue.append(doc)
                continue
            elif self.suppressed_docs_queue:
                self._print_suppression_msgs_queue()

            if isinstance(doc, OtherFile) and doc.is_valid_for_table and self.collect_other_files_to_tables:
                if self.new_names(doc):
                    if self.other_files_queue:
                        self._print_other_files_queue()  # Clear the queue before new biographical panel

                    self.print_characters_panel(self.new_names(doc))

                self.other_files_queue.append(doc)
                continue
            elif self.other_files_queue:
                self._print_other_files_queue()

            if isinstance(doc, Document):
                self.print_characters_panel(self.new_names(doc))
                doc.print()
                self._append_html_element(doc.to_html())
                self.printed_docs.append(doc)
            else:
                # TODO: this sucks
                console.print(doc)
                self.html_elements.append(doc.to_html())
                self.printed_file_displays.append(doc)

            if i % 100 == 0:
                timer.print_at_checkpoint(f"Printed {i} documents")

    def write_html(self, html_path: Path) -> None:
        """Write the collection of html elements to a file."""
        write_templated_html(self.html_elements, html_path)

    def _align_biographical_panel(self, panel: Panel) -> Align | None:
        return Align(Padding(panel, site_config.character_bio_padding), 'right')

    def _append_html_element(self, element: str) -> None:
        if self.last_bio_panel:
            element = '\n'.join([self.last_bio_panel, element])
            self.last_bio_panel = ''

        self.html_elements.append(div_class(element, 'doc_container'))

    def _biographical_panel(self, names: list[str]) -> Panel | None:
        """Panel showing biographical info for a list of names."""
        bios = [
            Text('', justify='right').append(PEOPLE_BIOS[name])
            for name in names if PEOPLE_BIOS.get(name)
        ]

        if not bios:
            return None

        return Panel(
            Group(*bios),
            border_style='dim',
            box=box.DOUBLE,
            expand=False,
            # padding=(0, 2),
            style='on gray7',
            title=Text(f"new names in next file", 'grey85 italic'),
            title_align='right',
        )

    def _html_so_far(self) -> str:
        return buffer_as_html(console, False)

    def _print_other_files_queue(self) -> None:
        has_printed_any_other_file_objs = any(isinstance(d, OtherFile) for d in self.printed_docs)

        if has_printed_any_other_file_objs:
            table_title = None
        else:
            table_title = OTHER_FILES_TABLE_MSG
            console.line()
            self.html_elements.append('<div style="height: 1em"></div>')

        table = OtherFile.files_preview_table(
            self.other_files_queue,
            title=table_title,
            title_justify='center',
        )

        console.print(Padding(table, (0, 0, 1, site_config.other_files_table_indent)))
        table_html = table_to_html(table, {'margin-left': to_em(site_config.other_files_table_indent)})
        self._append_html_element(table_html)
        self.printed_docs.extend(self.other_files_queue)
        self.other_files_queue = []

    def _print_suppression_msgs_queue(self) -> None:
        """Print any suppression messages."""
        for doc in self.suppressed_docs_queue:
            doc.print()

        msgs_panel = BasePanel(border_style='', text=[d.suppressed_txt for d in self.suppressed_docs_queue])
        self.html_elements.append(msgs_panel.to_div((site_config.info_indent, 0)))  # TODO: better indent handling
        self.suppressed_docs_queue = []
        console.line()



"""
Custom configurations for various files.
"""
from itertools import groupby

from epstein_files.documents.config.categories.emails import EMAILS_CONFIG
from epstein_files.documents.config.categories.imessage_logs import TEXTS_CONFIG
# These imports are actually used, but through a dynamic locals() reference
from epstein_files.documents.config.categories.academia import ACADEMIA_CFGS
from epstein_files.documents.config.categories.article import ARTICLE_CFGS
from epstein_files.documents.config.categories.arts import ARTS_CFGS
from epstein_files.documents.config.categories.book import BOOK_CFGS
from epstein_files.documents.config.categories.conference import CONFERENCE_CFGS
from epstein_files.documents.config.categories.crypto import CRYPTO_CFGS
from epstein_files.documents.config.categories.deposition import DEPOSITION_CFGS
from epstein_files.documents.config.categories.finance import FINANCE_CFGS
from epstein_files.documents.config.categories.girls import GIRLS_CFGS
from epstein_files.documents.config.categories.government import GOVERNMENT_CFGS
from epstein_files.documents.config.categories.junk import JUNK_CFGS
from epstein_files.documents.config.categories.legal import LEGAL_CFGS
from epstein_files.documents.config.categories.letter import LETTER_CFGS
from epstein_files.documents.config.categories.misc import MISC_CFGS
from epstein_files.documents.config.categories.money import MONEY_CFGS
from epstein_files.documents.config.categories.phone_bill import PHONE_BILL_CFGS
from epstein_files.documents.config.categories.politics import POLITICS_CFGS
from epstein_files.documents.config.categories.property import PROPERTY_CFGS
from epstein_files.documents.config.categories.reputation import REPUTATION_CFGS
from epstein_files.documents.config.categories.resume import RESUMÉ_CFGS
from epstein_files.documents.config.categories.skype_log import SKYPE_LOG_CFGS
from epstein_files.documents.config.categories.social import SOCIAL_CFGS, TWEET_CFGS
from epstein_files.documents.config.categories.text_msg import TEXT_MSG_CFGS
from epstein_files.documents.config.config_builder import victim_diary
from epstein_files.documents.config.doc_cfg import DocCfg, EmailCfg
from epstein_files.documents.documents.categories import CONSTANT_CATEGORIES, Interesting, Neutral
from epstein_files.people.names import *
from epstein_files.util.constant.strings import *
from epstein_files.util.helpers.string_helper import join_truthy, quote
from epstein_files.util.logging import logger

CFGS_SUFFIX = '_CFGS'
OTHER_FILES_PFX = 'OTHER_FILES_'
PARTICIPANTS_FIELD = 'Participants: field'
VALAR_FUND = f"{PETER_THIEL}'s {VALAR_VENTURES} fund"

HEADER_ABBREVIATIONS = {
    "AD": "Abu Dhabi",
    "Barak": f"former Israeli prime minister {EHUD_BARAK}",
    "Barrack": "Tom Barrack (Trump ally)",
    'BG, Bill': "Bill Gates",
    "Brock": 'Brock Pierce (crypto bro with a very sordid past)',
    "DB": "Deutsche Bank (maybe??)",
    "GRAT": "Grantor Retained Annuity Trust (tax shelter)",
    'HBJ, Jabor Y': "Sheikh Hamad bin Jassim (former Qatari prime minister)",
    'Jared': "Jared Kushner",
    'Jagland': 'Thorbjørn Jagland (former Norwegian prime minister)',
    'Joi': f"Joi Ito ({MIT_MEDIA_LAB}, MIT Digital Currency Initiative)",
    "Hoffenberg": f"{STEVEN_HOFFENBERG} (Epstein's ponzi scheme partner)",
    'KSA': "Kingdom of Saudi Arabia",
    'Kurz': 'Sebastian Kurz (former Austrian Chancellor)',
    'Kwok': "Chinese criminal Miles Kwok AKA Miles Guo AKA Guo Wengui",
    'LSJ': "Little St. James (Epstein's island)",
    'Madars': 'Madars Virza (co-founder of privacy crypto ZCash)',
    'Mapp': f'{KENNETH_E_MAPP} (Governor of {VIRGIN_ISLANDS})',
    'MBS': "Mohammed bin Salman Al Saud (Saudi ruler)",
    'MBZ': "Mohamed bin Zayed Al Nahyan (Emirates sheikh)",
    "Miro": MIROSLAV_LAJCAK,
    "Mooch": "Anthony 'The Mooch' Scaramucci (Skybridge crypto bro)",
    "NPA": 'non-prosecution agreement',
    "PA": PRINCE_ANDREW,
    "Terje": TERJE_ROD_LARSEN,
    "VI": f"U.S. {VIRGIN_ISLANDS}",
    "Woody": "Woody Allen",
}


################################################################################################
####################################### DOC CONFIGS ############################################
################################################################################################

DIARY_CFGS = [
    victim_diary('EFTA02731465', f"naming {JES_STALEY}, {TED_LEONSIS}, {GEORGE_VRADENBURG}, references tinkerbell"),
    victim_diary('EFTA02731420', f'naming {LARRY_SUMMERS}, {PRINCE_ANDREW}, Dan Snyder, {LEON_BLACK}, {TED_LEONSIS}'),
]

FLIGHT_LOG_CFGS = [
    DocCfg(id='022780'),
    DocCfg(id='022816'),
    DocCfg(id='EFTA00623147', author=DAVID_RODGERS, date='2016-06-30', date_uncertain='guess based on employment history'),
]

# Interesting / uninteresting
INTERESTING_EMAIL_IDS = [
    '033096',
    'EFTA02516264',
]

# These emails will be suppressed in the curated views
UNINTERESTING_EMAIL_IDS = [
    # Alan Dlugash
    'EFTA02367999',
    # Amir Taaki
    'EFTA01983108',
    # Austin Hill
    'EFTA01024046',
    'EFTA01010209',
    'EFTA00461202',
    'EFTA02229342',
    'EFTA01005712',
    'EFTA01005715',
    'EFTA02228570',
    # Bannon
    'EFTA02517956',
    '030710',
    '030954',
    '030956',
    '027585',  # texts
    '020440',
    '019338',
    '019343',
    '026304',
    '030372',
    '030408',
    '026310',
    '030736',
    '020443',
    '030983',
    '030958',
    '026477',
    # Ben Goertzel
    'EFTA01745739',
    # Brock
    'EFTA02174702',
    'EFTA00371547',
    'EFTA00405824',
    'EFTA00361736',
    'EFTA02160842',
    'EFTA00405795',
    'EFTA00362148',
    'EFTA00362163',
    'EFTA00997253',
    'EFTA00997251',
    'EFTA00362171',
    'EFTA00693620',
    'EFTA02092108',
    'EFTA00645400',
    'EFTA01002109',
    'EFTA02365466',
    'EFTA00709543',
    'EFTA00419736',  # visible elsewhere
    'EFTA01001754',  # visible elsewhere
    'EFTA00699275',
    'EFTA00864590',
    'EFTA00875181',
    # Christina Galbraith
    '031591',
    # David Stern
    'EFTA02507454',
    'EFTA02478704',
    'EFTA01862308',
    'EFTA02570707',
    '030724',
    # Ehud Barak
    '025878',
    '026354',
    '026618',
    # Epstein
    '030997',
    '033428',
    '030154',
    # Eric Roth
    '033386',
    # FBI
    'EFTA00005705',
    'EFTA00005716',
    'EFTA00005717',
    'EFTA00039971',  # Attached 302 is missing?
    # Ganbat
    'EFTA02469375',
    # Ghislaine
    'EFTA00582171',  # Visible in other files
    # Ian Osborne
    'EFTA01763771',
    'EFTA00932122',
    # Jabor
    '030786',
    '033011',
    # Jean Luc Brunel
    'EFTA01987855',  # TODO not entirely uninteresting...
    # Jeremy Rubin
    'EFTA00714127',
    # John Page
    '016693',
    # Joi Ito
    '029500',
    '029279',
    '029088',
    '029204',
    '029094',
    '029096',
    '028924',  # visible elsewhere
    '029224',  # visible elsewhere
    '029429',
    '029587',
    '029237',
    '029499',
    'EFTA02363524',  # visible in EFTA00676383
    'EFTA00649516',
    'EFTA02593836',
    'EFTA00697880',
    'EFTA02647229',
    'EFTA02238841',
    # Karyna Shuliak
    'EFTA02318365',
    # Krassner
    '033345',
    # Leon Black related
    '023208_24',
    '023208_26',
    # Lesley?
    'EFTA02229402',
    'EFTA02229659',
    'EFTA01987383',
    'EFTA02067872',
    # Maria Prusakova
    'EFTA01772533',
    'EFTA01740489',
    # Peter Thiel
    'EFTA01003336',
    # Philip Rosedale
    'EFTA01903448',
    'EFTA01899565',
    # Ramsey Elkholy
    'EFTA02438522',  # visible elsewhere
    # Rasseck Bourgi
    'EFTA01990389',  # visible in EFTA01988194
    # Renata Bolotova
    'EFTA01903041',
    'EFTA01969322',
    'EFTA01035614',
    'EFTA02719248',
    'EFTA02696764',
    'EFTA00927986',
    'EFTA00667874',
    # Svetlana
    'EFTA01772677',
    # Tyler Shears
    'EFTA02511281',
    # USANYS / DOJ
    'EFTA00039799',
    'EFTA02730469',
    'EFTA02730471',
    'EFTA02731662',
    'EFTA02731648',
    'EFTA02731473',
    'EFTA02731735',
    'EFTA02731734',
    'EFTA02731628',
    'EFTA00040144',
    'EFTA00040105',
    'EFTA00040118',
    'EFTA00040124',
    'EFTA02730468',
    'EFTA00040141',
    'EFTA02730485',
    'EFTA02731526',
    'EFTA00039802',
    'EFTA00039867',
    'EFTA00039995',
    'EFTA00039981',
    'EFTA00104945',
    'EFTA00039893',
    'EFTA02385456',
    'EFTA02730481',
    'EFTA02730483',
    # TODO: These have UNKNOWN recipient so they currently get printed but we should configure it so they don't
    'EFTA00039894',
    'EFTA00039878',
    # Valdson
    'EFTA02303690',
    # Vincenzo Iozzo
    '033280',
    'EFTA02624738',
    'EFTA01751416',  # Visible in EFTA02584771
    # Wolff
    '021120',
    # Unknown
    '030992',
    '032213',
    '029206',
    '031822',
    '031705',
    '030768',
    '026659',
    '032951',
    '023062',
    '030324',
    '031990',
    '024930',
    '029982',
    '022187',
    '033486',
    '029446',
    '019873',
    '030823',  # "little hodiaki"
    '027009',
    '026273',
    'EFTA02431535',  # visible in EFTA00888467
    'EFTA00363992',  # car rental
    'EFTA02187735',  # housekeeping
]

# Not uninteresting enough to be permanently marked as such but not good enough for --output-chrono
NOT_CHRONOLOGICAL_VIEW_IDS = [cfg.id for cfg in FLIGHT_LOG_CFGS] + [
    # Karim Wade
    'EFTA01063216',  # canceled trip
    'EFTA01738267',  # canceled trip
    # Ehud barak
    '032336',
    '033338',
    'EFTA00582504',
    '024432',
    '019352',
    'EFTA02053321',  # Ike Groff
    'EFTA00915298',
    'EFTA02423635',
    'EFTA00007781',
    'EFTA00770066',
    '016692',
    'EFTA01823635',
    'EFTA00747181',
    'EFTA00758140',
    'EFTA01979689',
    'EFTA00878255',
    'EFTA00766770',
    'EFTA00384774',
    'EFTA00892654',  # Mandelson
    'EFTA02437992',  # Ben Goertzel
    'EFTA01613143',  # Melanie Walker
    'EFTA02144766',  # Zorro Ranch
    'EFTA02308934',  # Yoed Nir
    'EFTA02009735',  # Boris email about Regina Dugan
    'EFTA00719151',  # Boris Regina TED
    '030609',
    '022247',
    '030095',
    '022234',
    '030470',
    '030222',
    # Christina Galbraith
    'EFTA00941810'
    # '031826',
    '031826',
    '019446',  #Haiti jacmel
    # Jean Luc visa etc
    'EFTA02515416',
    'EFTA02516675',
    '011908_6',
    # Jabor
    # Joi Ito
    'EFTA01058627',
    # Deepak
    'EFTA02342230',
    # Misc
    'EFTA00329334',
    # Renata Bolotova
    'EFTA00667441',
    # Tyler Shears
    'EFTA02372294',
    '028965',
    # --- #
    'EFTA02048499',
    'EFTA00322570',
    'EFTA02001536',
    'EFTA01773417',
    'EFTA01987283',
    'EFTA01778423',
    'EFTA01798022',
    'EFTA00845330',
    'EFTA00840747',
    'EFTA02475571',
    '030807',
    '032842',
    'EFTA00434111',
    'EFTA00432352',
    '026583',
    '031986',
    '026584',
    '029973',
    '031038',
    '028760',
    '029976',
    '026320',
    '031169',
    'EFTA01757037',
    '030211',
    '031964',
    '027032',
    '027126',
    '030881',
    'EFTA00373948',
    'EFTA00659818',
    'EFTA01931339',
    'EFTA01923844',
    'EFTA00998902',
    '011908_4',
    '030874',
    'EFTA02396796',
    '031320',
    '026254',
    '026473',
    '026474',
    '025517',
    '025520',
    '030721',
    '030785',
    '029497',
    '030795',
    'EFTA00848188',
    'EFTA02465832',
    'EFTA02430577',
    '030444',
    '030788',
    '020815',
    '021106',
    '027044',
    '025429',
    '027583',
    '025426',
    '025423',
    '027707',
    '019302',
    '019305',
    '019334',
    '019308',
    '019312',
    '026311',
    '027067',
    '030966',
    '030967',
    '030986',
    '029583',
    '017827',
    # '024185', # UN
]


# Build config list by combining [BLAH]_CFGS variables and setting category to [BLAH] for each
CATEGORY_CONFIGS: list[DocCfg] = []

for category in CONSTANT_CATEGORIES:
    var_name = f"{category.upper()}{CFGS_SUFFIX}"

    if var_name not in locals():
        logger.warning(f"Document config variable '{var_name}' is not defined!")
        continue

    category_cfgs = locals()[var_name]
    DocCfg.set_categories(category_cfgs, category)
    CATEGORY_CONFIGS.extend(category_cfgs)

ALL_CONFIGS = CATEGORY_CONFIGS + EMAILS_CONFIG + TEXTS_CONFIG
EmailCfg.create_or_set_prop(INTERESTING_EMAIL_IDS, ALL_CONFIGS, 'is_interesting', True)
EmailCfg.create_or_set_prop(UNINTERESTING_EMAIL_IDS, ALL_CONFIGS, 'is_interesting', False)
EmailCfg.create_or_set_prop(NOT_CHRONOLOGICAL_VIEW_IDS, ALL_CONFIGS, 'is_in_chrono', False)
CONFIGS_BY_ID = {cfg.id: cfg for cfg in ALL_CONFIGS}

# Add synthetic Cfg objects for duplicate docs with same props as the DocCfg they are a duplicate of
for cfg in ALL_CONFIGS:
    for dupe_cfg in cfg.duplicate_cfgs():
        CONFIGS_BY_ID[dupe_cfg.id] = dupe_cfg


# Collect special docs to show with special people
SHOW_WITH_DOCS = {
    id: list(cfgs)
    for id, cfgs in groupby(ALL_CONFIGS, lambda cfg: cfg.show_with_name)
    if id
}


def check_no_overlapping_configs():
    encountered_file_ids = set()

    for cfg in ALL_CONFIGS:
        if cfg.duplicate_of_id:
            assert cfg.duplicate_of_id != cfg.id, f"Bad config! {cfg}"

        if cfg.id in encountered_file_ids:
            raise RuntimeError(f"{cfg.id} configured twice!")

        encountered_file_ids.add(cfg.id)


check_no_overlapping_configs()



from rich.align import Align
from rich.console import Group
from rich import box
from rich.table import Table
from rich.text import Text
from typing import Sequence

from epstein_files.util.env import args

LINES_WIDTH = 53


def build_demi_table(left_col_msg: str, right_col_list: Sequence[str | Text]) -> Group:
    """Table with only one cell on left and only top, bottom, and middle lines."""
    container = Table(box=box.ROUNDED, show_header=False, collapse_padding=True, show_edge=False, style='gray23')
    container.add_column('msg', width=19, style='dim', justify='right')
    container.add_column('link', justify='left')
    inner_list = Table(show_header=False, collapse_padding=True, show_edge=False, style='gray23')
    inner_list.add_column('link', justify='left')

    for element in right_col_list:
        inner_list.add_row(element)

    container.add_row(left_col_msg, inner_list)
    line_width = min(args.width or LINES_WIDTH, LINES_WIDTH) - 3
    top_bottom_border = Text('-' * line_width, style='grey15')

    pieces = [
        top_bottom_border,
        container,
        top_bottom_border
    ]

    return Group(*[Align.center(p) for p in pieces])



"""
Finance related documents not specifically about Epstein's money.
"""
from epstein_files.documents.documents.categories import Interesting, Neutral
from epstein_files.documents.config.doc_cfg import GOLDMAN_INVESTMENT_MGMT, DocCfg, EmailCfg
from epstein_files.people.names import *
from epstein_files.util.constant.strings import *

BOFA_WEALTH_MGMT = f'{BOFA} Wealth Management'
DEUTSCHE_BANK_TAX_TOPICS = f'{DEUTSCHE_BANK} Wealth Management Tax Topics'
JP_MORGAN_EYE_ON_THE_MARKET = f"Eye On The Market"
UBS_CIO_REPORT = 'CIO Monthly Extended report'


# All authors of documents in this category will be marked uninteresting
FINANCE_CFGS = [
    DocCfg(id='024631', author='Ackrell Capital', description=f"Cannabis Investment Report 2018", is_interesting=True),
    DocCfg(id='016111', author=BOFA_MERRILL, description=f"GEMs Paper #26 Saudi Arabia: beyond oil but not so fast", date='2016-06-30'),
    DocCfg(id='010609', author=BOFA_MERRILL, description=f"Liquid Insight Trump's effect on MXN", date='2016-09-22'),
    DocCfg(id='025978', author=BOFA_MERRILL, description=f"Understanding when risk parity risk Increases", date='2016-08-09'),
    DocCfg(id='014404', author=BOFA_MERRILL, description=f'Japan Investment Strategy Report', date='2016-11-18'),
    DocCfg(id='014410', author=BOFA_MERRILL, description=f'Japan Investment Strategy Report', date='2016-11-18'),
    DocCfg(id='014424', author=BOFA_MERRILL, description=f"Japan Macro Watch", date='2016-11-14'),
    DocCfg(id='014731', author=BOFA_MERRILL, description=f"Global Rates, FX & EM 2017 Year Ahead", date='2016-11-16'),
    DocCfg(id='014432', author=BOFA_MERRILL, description=f"Global Cross Asset Strategy - Year Ahead The Trump inflection", date='2016-11-30'),
    DocCfg(id='014460', author=BOFA_MERRILL, description=f"European Equity Strategy 2017", date='2016-12-01'),
    DocCfg(id='014972', author=BOFA_MERRILL, description=f"Global Equity Volatility Insights", date='2017-06-20'),
    DocCfg(id='014622', author=BOFA_MERRILL, description=f"Top 10 US Ideas Quarterly", date='2017-01-03'),
    DocCfg(id='023069', author=BOFA_MERRILL, description=f"Equity Strategy Focus Point Death and Taxes", date='2017-01-29'),
    DocCfg(id='014721', author=BOFA_MERRILL, description=f"Cause and Effect Fade the Trump Risk Premium", date='2017-02-13'),
    DocCfg(id='014887', author=BOFA_MERRILL, description=f"Internet / e-Commerce", date='2017-04-06'),
    DocCfg(id='014873', author=BOFA_MERRILL, description=f"Hess Corp", date='2017-04-11'),
    DocCfg(id='023575', author=BOFA_MERRILL, description=f"Global Equity Volatility Insights", date='2017-06-01'),
    DocCfg(id='014518', author=BOFA_WEALTH_MGMT, description=f'tax alert', date='2016-05-02'),
    DocCfg(id='029438', author=BOFA_WEALTH_MGMT, description=f'tax report', date='2018-01-02'),
    DocCfg(id='026668', author="Boothbay Fund Management", description=f"2016-Q4 earnings report signed by Ari Glass", is_interesting=True),
    DocCfg(id='024302', author='Carvana', description=f"form 14A SEC filing proxy statement", date='2019-04-23'),
    DocCfg(id='029305', author='CCH Tax', description=f"Briefing on end of Defense of Marriage Act", date='2013-06-27'),
    DocCfg(id='026794', author=DEUTSCHE_BANK, description=f"Global Political and Regulatory Risk in 2015/2016"),
    DocCfg(id='022361', author=DEUTSCHE_BANK_TAX_TOPICS, date='2013-05-01', attached_to_email_id='022359'),
    DocCfg(id='022325', author=DEUTSCHE_BANK_TAX_TOPICS, date='2013-12-20'),
    DocCfg(id='022330', author=DEUTSCHE_BANK_TAX_TOPICS, date='2013-12-20', description='table of contents'),
    DocCfg(id='019440', author=DEUTSCHE_BANK_TAX_TOPICS, date='2014-01-29'),
    DocCfg(id='024202', author=ELECTRON_CAPITAL_PARTNERS, description=f"Global Utility White Paper", date='2013-03-08'),
    DocCfg(id='022372', author='Ernst & Young', date='2016-11-09', description=f'2016 election report'),
    DocCfg(id='014532', author=GOLDMAN_INVESTMENT_MGMT, description=f"Outlook - Half Full", date='2017-01-01'),
    DocCfg(
        id='026909',
        attached_to_email_id='026893',
        author=GOLDMAN_INVESTMENT_MGMT,
        description=f"The Unsteady Undertow Commands the Seas (Temporarily)",
        date='2018-10-14',
    ),
    DocCfg(id='026944', author=GOLDMAN_INVESTMENT_MGMT, description=f"Risk of a US-Iran Military Conflict", date='2019-05-23'),
    DocCfg(id='018804', author='Integra Realty Resources', description=f"appraisal of going concern for IGY American Yacht Harbor Marina in {VIRGIN_ISLANDS}"),
    DocCfg(id='026679', author='Invesco', description=f"Global Sovereign Asset Management Study 2017"),
    DocCfg(
        id='033220',
        author='Joseph G. Carson',
        description=f"short economic report on defense spending under Trump",
        is_interesting=True,
    ),
    DocCfg(id='026572', author=JP_MORGAN, description=f"Global Asset Allocation report", date='2012-11-09'),
    DocCfg(id='030848', author=JP_MORGAN, description=f"Global Asset Allocation report", date='2013-03-28'),
    DocCfg(id='030840', author=JP_MORGAN, description=f"Market Thoughts"),
    DocCfg(id='022350', author=JP_MORGAN, description=f"tax efficiency of Intentionally Defective Grantor Trusts (IDGT)"),
    DocCfg(id='025242', author=JP_MORGAN, description=JP_MORGAN_EYE_ON_THE_MARKET, date='2012-04-09'),
    DocCfg(id='030010', author=JP_MORGAN, description=JP_MORGAN_EYE_ON_THE_MARKET, attached_to_email_id='030006', date='2011-06-14'),
    DocCfg(id='030808', author=JP_MORGAN, description=JP_MORGAN_EYE_ON_THE_MARKET, date='2011-07-11'),
    DocCfg(id='025221', author=JP_MORGAN, description=JP_MORGAN_EYE_ON_THE_MARKET, date='2011-07-25'),
    DocCfg(id='025229', author=JP_MORGAN, description=JP_MORGAN_EYE_ON_THE_MARKET, date='2011-08-04'),
    DocCfg(id='030814', author=JP_MORGAN, description=JP_MORGAN_EYE_ON_THE_MARKET, date='2011-11-21'),
    DocCfg(id='024132', author=JP_MORGAN, description=JP_MORGAN_EYE_ON_THE_MARKET, date='2012-03-15'),
    DocCfg(id='024194', author=JP_MORGAN, description=JP_MORGAN_EYE_ON_THE_MARKET, date='2012-10-22'),
    DocCfg(id='025296', author='Laffer Associates', description=f'report predicting Trump win', date='2016-07-06'),
    DocCfg(
        id='020824',
        author='Mary Meeker',
        date='2011-02-01',
        description=f"USA Inc: A Basic Summary of America's Financial Statements compiled",
    ),
    DocCfg(id='025551', author='Morgan Stanley', description=f'report about alternative asset managers', date='2018-01-30'),
    DocCfg(id='019856', author='Sadis Goldberg LLP', description=f"report on SCOTUS ruling about insider trading", is_interesting=True),
    DocCfg(id='025763', author='S&P', description=f"Economic Research: How Increasing Income Inequality Is Dampening U.S. Growth", date='2014-08-05'),
    DocCfg(id='024135', author=UBS, description=UBS_CIO_REPORT, date='2012-06-29'),
    DocCfg(id='025247', author=UBS, description=UBS_CIO_REPORT, date='2012-10-25'),
    DocCfg(id='026584', description="article about tax implications of disregarded entities", date='2009-07-01', is_interesting=True),
    DocCfg(id='024817', description="Cowen's CBD / Cannabis report", date='2019-02-25', is_interesting=True),
    DocCfg(
        id='012048',
        category=Neutral.PRESSER,
        description=f"Rockefeller Partners with Gregory J. Fleming to Create Independent Financial Services Firm and other articles",
        is_interesting=False,
    ),

    # DOJ
    DocCfg(id='EFTA00286949', author='Nathaniel August', description='Mangrove Partners valuation report', date='2016-11-03'),
    DocCfg(id='EFTA01078403', author=ATORUS, description='investment adviser uniform application', date='2014-05-14'),
    DocCfg(id='EFTA02690885', author=ATORUS, description='pitch deck'),
    DocCfg(id='EFTA00593276', author=EDMOND_DE_ROTHSCHILD, description='org chart', is_interesting=True),
    # Prop trading / HFT
    DocCfg(id='EFTA00611806', author='Adam Campbell', description='CEA, LLC High Liquidity Trading Program presentation'),
    DocCfg(id='EFTA01088824', author='Boothbay', description='Absolute Return Strategies Fund pitch deck'),
    DocCfg(id='EFTA00556664', author='Qarmin', description='high frequency trading pitch deck'),
]



"""
Helpers for dealing with various kinds of data.
"""
import itertools
import re
from datetime import datetime, timezone
from dateutil import tz
from typing import Any, Mapping, Sequence, TypeVar

from epstein_files.people import names
from epstein_files.util.env import args
from epstein_files.util.logging import logger

T = TypeVar('T')

ISO_DATE_REGEX = re.compile(r'\d{4}-\d{2}(-\d{2})?')
CONSTANT_VAR_REGEX = re.compile(r"^[A-Z_]+$")
ALL_NAMES = [v for k, v in vars(names).items() if isinstance(v, str) and CONSTANT_VAR_REGEX.match(k)]

AMERICAN_DATE_FORMAT = r"%m/%d/%y %I:%M:%S %p"
AMERICAN_TIME_REGEX = re.compile(r"(\d{1,2}/\d{1,2}/\d{2,4}\s+\d{1,2}:\d{2}(?::\d{2})?\s*(?:AM|PM)?)")
PACIFIC_TZ = tz.gettz("America/Los_Angeles")
TIMEZONE_INFO = {"PDT": PACIFIC_TZ, "PST": PACIFIC_TZ}  # Suppresses annoying warnings from parse() calls

all_elements_same = lambda _list: len(_list) == 0 or all(x == _list[0] for x in _list)
date_str = lambda dt: dt.isoformat()[0:10] if dt else None
dict_key_list = lambda d: [k for k in d.keys()]
escape_double_quotes = lambda text: text.replace('"', r'\"')
escape_single_quotes = lambda text: text.replace("'", r"\'")
days_between = lambda dt1, dt2: (dt2 - dt1).days + 1
days_between_str = lambda dt1, dt2: f"{days_between(dt1, dt2)} day" + ('s' if days_between(dt1, dt2) > 1 else '')
remove_zero_time = lambda dt: dt.isoformat().removesuffix('T00:00:00')
uniquify = lambda _list: list(set(_list))
without_falsey = lambda _list: [e for e in _list if e]


def constantize_names(s: str) -> str:
    """Replace occurences of 'Jeffrey Epstein' with '{JEFFREY_EPSTEIN}' etc."""
    for name in ALL_NAMES:
        s = s.replace(name, f"{{{names.constantize_name(name)}}}")

    return s


def dict_sets_to_lists(d: dict[str, set]) -> dict[str, list]:
    """Turn the `set` values in the dict into sorted lists."""
    return {k: sorted(list(v)) for k, v in d.items()}


def flatten(_list: Sequence[list[T] | set[T]]) -> list[T]:
    if not _list:
        return []
    elif all(isinstance(element, set) for element in _list):
        return list(set().union(*_list))
    else:
        return list(itertools.chain.from_iterable(_list))


def json_safe(d: dict) -> dict:
    return {
        'None' if k is None else k: v.isoformat() if isinstance(v, datetime) else v
        for k,v in d.items()
    }


def listify(listlike) -> list:
    """Create a list of 'listlike'. Returns empty list if 'listlike' is None or empty string."""
    if isinstance(listlike, list):
        return listlike
    elif listlike is None:
        return [None]
    elif listlike:
        return [listlike]
    else:
        return []


def ordinal_str(n: int) -> str:
    if 11 <= (n % 100) <= 13:
        suffix = 'th'
    else:
        suffix = {1: 'st', 2: 'nd', 3: 'rd'}.get(n % 10, 'th')

    return str(n) + suffix


def patternize(_pattern: str | re.Pattern) -> re.Pattern:
    return _pattern if isinstance(_pattern, re.Pattern) else re.compile(fr"({_pattern})", re.IGNORECASE)


def prefix_keys(prefix: str, _dict: dict[str, T], sep='.') -> dict[str, T]:
    """Add `prefix` to the front of all the keys in `_dict`."""
    return {f"{prefix}{sep}{k}": v for k, v in _dict.items()}


def remove_timezone(timestamp: datetime) -> datetime:
    if timestamp.tzinfo:
        timestamp = timestamp.astimezone(timezone.utc).replace(tzinfo=None)

    return timestamp


def sort_dict(d: dict[str | None, int] | dict[str, int]) -> list[tuple[str | None, int]]:
    alpha_key = lambda kv: (kv[0] or '').lower()

    try:
        sort_key = alpha_key if args.sort_alphabetical else lambda kv: [-kv[1], alpha_key(kv)]
        return sorted(d.items(), key=sort_key)
    except TypeError as e:
        alpha_key = lambda kv: kv[0] or ''
        return sorted(d.items(), key=lambda kv: f"Z{alpha_key(kv)}" if '.' in (kv[0] or '') else alpha_key(kv))


def sort_dict_by_keys(d: Mapping[names.Name, T]) -> dict[names.Name, T]:
    return {k: d[k] for k in names.sort_names(dict_key_list(d))}


def uniq_sorted(_list: list[T]) -> list[T]:
    return sorted(uniquify(_list))


def uniquify(_list: list[T]) -> list[T]:
    return list(set(_list))



"""
Crypto and virtual currency / fintech related files
"""
from epstein_files.documents.config.categories.finance import GOLDMAN_INVESTMENT_MGMT
from epstein_files.documents.config.config_builder import whistleblower_cfg
from epstein_files.documents.config.doc_cfg import DEFAULT_TRUNCATE_TO, NO_TRUNCATE, SHORT_TRUNCATE_TO, DocCfg, EmailCfg
from epstein_files.people.names import *
from epstein_files.util.constant.strings import *
from epstein_files.util.helpers.string_helper import join_truthy, quote

BEN_LAWSKY_NYDFS = f'head of NY Dept of Financial Services {BEN_LAWSKY}'
HONEYCOMB_FUND = 'Honeycomb Asset Management (Spotify and Tencent investors)'
KYARA_FUND = f"Epstein crypto fund {KYARA_INVESTMENT}"
RIOT_BLOCKCHAIN_DESCRIPTION = 'RIOT Blockchain (FKA "Bioptix") is a sketchy bitcoin miner in Texas'
VALAR_FUND = f"{PETER_THIEL}'s {VALAR_VENTURES} fund"


# This category is automatically 'interesting', see OtherFile class
CRYPTO_CFGS = [
    DocCfg(id='031743', description=f'a few pages describing the internet as a "New Nation State" (Network State?)'),
    DocCfg(
        id='025663',
        author=GOLDMAN_INVESTMENT_MGMT,
        description=f"An Overview of the Current State of Cryptocurrencies and Blockchain",
        date='2017-11-15',
        duplicate_ids=['EFTA00793120'],
    ),
    # Alternative currencies
    EmailCfg(
        id='EFTA02021870',
        author=BARNABY_MARSH,
        date='2012-01-29 11:15:00',
        highlight_quote="I would like to see you alone to discuss alternative currencies",
        recipients=[JEFFREY_EPSTEIN],
    ),
    EmailCfg(
        id='EFTA00941810',
        author_reason='"chris" in reply, "cell:/email" in signature',
        recipients=[CHRISTINA_GALBRAITH],
        truncate_to=1_000,
    ),
    EmailCfg(id='EFTA01852443'),
    EmailCfg(id='EFTA01754301'),
    EmailCfg(id='EFTA02361956', author=LINDA_STONE, author_reason='signature'),
    EmailCfg(id='EFTA02575430', description='alternative currencies group'),
    EmailCfg(id='EFTA00660838', description=f"Epstein says his trip to Ivory Coast was amazing", truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA02391375', highlight_quote="Have you heard of the alternative currency"),
    # Bannon
    EmailCfg(id='030711', description='Epstein says "we can discuss michael and his coins", unclear what that means'),
    EmailCfg(id='026260', comment='Bannon cripto coin issues'),
    EmailCfg(id='026258', description=f'Bannon suggests {JEFFREY_WERNICK}, Parler COO and man affiliated with many shady crypto businesses, can help'),
    EmailCfg(id='030781', description="Esptein says the crypto coin issues they're having are 'US based'"),
    EmailCfg(id='026255', description='Epstein requests help with "coin issues" and "prohibitions foreign donor" (sic)', is_interesting=True),
    EmailCfg(id='EFTA02521522'),
    EmailCfg(id='EFTA02524752', description=f'{STEVE_BANNON} "is thinking of combining crypto with his politcial movement"', show_with_name=STEVE_BANNON),
    EmailCfg(id='EFTA01020383', description='Epstein emails Bannon a quote about the use of crypto to evade capital controls'),
    EmailCfg(id='EFTA00874793', description=f'Bannon says that they need to stop EU regulation of crypto scams "dead in its tracks"', is_interesting=True),
    EmailCfg(id='EFTA02517572', description=f'Bannon says that they need to stop EU regulation of crypto scams "dead in its tracks"'),
    EmailCfg(id='EFTA00881711', description=f"Bannon and Epstein discuss bitcoin and crypto donations", is_interesting=True),
    EmailCfg(id='EFTA00881759', description=f"Epstein and {STEVE_BANNON} discuss collection donations in bitcoin", is_interesting=True),
    # Bill Gates
    EmailCfg(id='EFTA02639760', description='Epstein pushing cryptocurrency on Bill Gates'),
    # Bitmain
    EmailCfg(id='EFTA01004576'),
    # Blockstream
    DocCfg(id='EFTA00797613', description='Alphabit crypto fund pitch deck', attached_to_email_id='EFTA00955063'),
    EmailCfg(id='EFTA02229342', author=AUSTIN_HILL),
    EmailCfg(
        id='EFTA02406818',
        author=AUSTIN_HILL,
        author_reason='unique email signature',
        description=f"{AUSTIN_HILL} asks Epstein about how the Treasury Dept. will treat dollar backed stablecoins like Tether",
    ),
    EmailCfg(
        id='EFTA00955063',
        description=f'{AUSTIN_HILL} explains that some crypto funds are engaged in fraudulent wash trading (the attached Excel sheet listing those funds has not been found)',
        is_interesting=True,
    ),
    EmailCfg(
        id='EFTA00673841',
        description="@pmarca is Marc Andreessen (a16z)",
        highlight_quote='business, Bitcoin, personal & esoteric mind games that we play',
        truncate_to=NO_TRUNCATE,
    ),
    EmailCfg(id='EFTA02529170', description=f"Blockstream set up a special company to invest in an Asian ride hailing service"),
    EmailCfg(id='EFTA01917402', description=f'discussion of Epstein investment in Blockstream funneled through {JOI_ITO}', is_interesting=True),
    EmailCfg(id='EFTA00999897', description=f"Epstein claims he's a fan of {ADAM_BACK} (who has recently claimed he had nothing to do with Epstein)"),
    EmailCfg(id='EFTA01915883', description=f"{JOI_ITO} doesn't trust {AUSTIN_HILL} (for good reason)"),
    EmailCfg(id='EFTA00988395', description=f'"Andy Back" is probably Blockstream bitcoin dev {ADAM_BACK}', duplicate_ids=['EFTA01928856']),
    EmailCfg(id='EFTA01925969', description=f"{AUSTIN_HILL} and {ADAM_BACK} and some women plan a trip to Epstein's island", is_interesting=True),
    EmailCfg(id='EFTA00929418', description=f"Epstein forwards information from {AUSTIN_HILL} about fraudulent crypto trading operations to {PAUL_BARRETT}"),
    EmailCfg(id='EFTA01010209', description=f"{AUSTIN_HILL} calls {AMIR_TAAKI} \"a bit crazy\'"),
    EmailCfg(id='EFTA01788532', description=f"{AUSTIN_HILL} asking about Jeffrey Katzenberg's contact info"),
    EmailCfg(id='EFTA02515798', recipients=[LINDA_STONE], description='Epstein confirms he invested in Blockstream'),  # response to EFTA02515756
    # Blockchain Capital / Brock Pierce / Crypto Currency Partners
    DocCfg(id='024271', author=BLOCKCHAIN_CAPITAL, description="pitch deck", date='2015-10-01'),
    EmailCfg(
        id='EFTA00689503',
        description=f"{BROCK_PIERCE} posts review of infamous Russian criminal crypto exchange BTC-e to LinkedIn",
        is_interesting=True,
        show_with_name=BROCK_PIERCE,
        # truncate_to=(1_400, 3_050),  # TODO: enable excerpts in emails
    ),
    DocCfg(id='EFTA01088644', author=BLOCKCHAIN_CAPITAL, description="pitch deck", date='2015-10-01'),
    DocCfg(id='EFTA00604942', author=BLOCKCHAIN_CAPITAL, description="Investor Update", date='2015-10-01'),
    DocCfg(id='EFTA01089535', author=CRYPTO_CURRENCY_PARTNERS_II, description=f"investor report", date='2014-10-01'),
    DocCfg(id='EFTA01089506', author=CRYPTO_CURRENCY_PARTNERS_II, description=f"investor report", date='2014-11-01'),
    DocCfg(id='EFTA01089500', author=CRYPTO_CURRENCY_PARTNERS_II, description=f"investor report", date='2015-01-01'),
    DocCfg(id='EFTA00621239', author=CRYPTO_CURRENCY_PARTNERS_II, description=f"investor report", date='2014-10-01'),
    DocCfg(id='EFTA01089526', author=CRYPTO_CURRENCY_PARTNERS_II, description=f"investor report", date='2014-08-01', attached_to_email_id='EFTA00997803'),
    DocCfg(id='EFTA01089485', author=CRYPTO_CURRENCY_PARTNERS_II, description=f"investor report", date='2014-09-01', attached_to_email_id='EFTA00693621'),
    DocCfg(id='EFTA01093509', author=CRYPTO_CURRENCY_PARTNERS_II, description=f"partnership draft", truncate_to=DEFAULT_TRUNCATE_TO * 2),
    DocCfg(id='EFTA01093444', author=CRYPTO_CURRENCY_PARTNERS_II, description='application form', attached_to_email_id='EFTA00626220'),  #EFTA02708994
    DocCfg(id='EFTA01093471', author=CRYPTO_CURRENCY_PARTNERS_II, description='operating agreement', attached_to_email_id='EFTA00626220'),
    DocCfg(id='EFTA01126582', author=EXPRESSCOIN, description=f"pitch deck for {BROCK_PIERCE} co.", attached_to_email_id='EFTA00990186'),
    DocCfg(id='EFTA01125964', author=EXPRESSCOIN, description=f"pitch deck for {BROCK_PIERCE} co.", attached_to_email_id='EFTA00985076'),
    EmailCfg(id='EFTA02465656', description=f"{BROCK_PIERCE} describes investment in video game items company"),
    EmailCfg(id='EFTA00673123', description=f"{BROCK_PIERCE} and Epstein invested in Coinbase which would turn out to be extremely lucrative", duplicate_ids=['EFTA02371875']),
    EmailCfg(id='EFTA01051074', description=f'{BROCK_PIERCE} arranging meeting between Bannon and Epstein (in Doha?)'),
    EmailCfg(id='EFTA02361144', description=f'{BROCK_PIERCE} brings Epstein an opportunity to invest in Blockstream, bitcoin developer {ADAM_BACK}\'s company'),
    EmailCfg(id='EFTA00999392', description=f"{BROCK_PIERCE} discussing a possibly sham marriage to someone named Sue for immigration purposes"),
    EmailCfg(id='EFTA02514539', description=f"{BROCK_PIERCE} has tax problems (hasn't filed a tax return in 5 years)", is_interesting=True),
    EmailCfg(id='EFTA00658457', description=f'{BROCK_PIERCE} says they will not invest in Bitfury', truncate_to=700),
    EmailCfg(id='EFTA00998595', description=f'{BROCK_PIERCE} and Epstein discuss structure of crypto investments'),
    EmailCfg(id='EFTA01000882', description=f"{BROCK_PIERCE} asks for Epstein's permission to disclose his name to Coinbase team", is_interesting=True),
    EmailCfg(id='EFTA00664424', description=f"{BROCK_PIERCE} suggests {LARRY_SUMMERS} could get involved with Tether and Noble Bank"),
    EmailCfg(id='EFTA01139067', description=f'Epstein tells Farkas he has not invested in {CRYPTO_CURRENCY_PARTNERS_II}', is_interesting=True),
    EmailCfg(id='EFTA00752383', description=f"Mindshift conference attendee list including {BROCK_PIERCE}", is_interesting=True),
    EmailCfg(id='EFTA00626220', description=f'setting up "sidecar fund" with {BROCK_PIERCE} / {CRYPTO_CURRENCY_PARTNERS_II}'),
    # Boris Nikolic
    EmailCfg(id='EFTA00844780'),
    EmailCfg(id='EFTA01770220'),
    EmailCfg(id='EFTA02037504'),
    EmailCfg(id='EFTA00972689', description=f"Epstein and Nikolic ruminate on the Silk Road bust", truncate_to=NO_TRUNCATE),
    EmailCfg(
        id='EFTA02436105',
        description="one of Epstein's goals in life was to make a lot of money in virtual currency",
        is_interesting=True,
    ),
    # Coinbase
    DocCfg(
        id='EFTA01082451',
        author=BLOCKCHAIN_CAPITAL,
        date='2014-11-10',
        description=f'/ {COINBASE} nondisclosure agreement possibly regarding Epstein investment',
        is_interesting=True,
        truncate_to=2000,
    ),
    DocCfg(id='EFTA01401059', description=f"{DEUTSCHE_BANK} wire transfer receipt for Epstein's investment in {BLOCKSTREAM}"),
    DocCfg(id='EFTA01092555', author=COINBASE, description='certificate of incorporation', date='2013-12-12', truncate_to=SHORT_TRUNCATE_TO),
    DocCfg(id='EFTA01120975', author=COINBASE, description='Series B stock purchase', date='2013-12-12', truncate_to=DEFAULT_TRUNCATE_TO),
    DocCfg(id='EFTA01121035', author=COINBASE, description='Series C preferred stock purchase', date='2014-10-24'),
    EmailCfg(id='EFTA02606117', comment='Coinbase valuation'),
    EmailCfg(id='EFTA02608112', comment='Coinbase valuation'),
    EmailCfg(id='EFTA00901772', description=f"Epstein's investment in {COINBASE} would earn him most of $100 million", is_interesting=True),
    EmailCfg(id='EFTA00900908', description="negotiation of repurchase of part of Epstein's stake in Coinbase"),
    # Crypto PR Lab
    DocCfg(id='EFTA00295126', author=CRYPTO_PR_LAB, description="accounting statement", show_full_panel=True),
    DocCfg(id='EFTA01299820', author=CRYPTO_PR_LAB, description="bank transfer", date='2018-12-18'),
    DocCfg(id='EFTA00309271', author=CRYPTO_PR_LAB, description='financial statement'),
    DocCfg(id='EFTA01613759', author=CRYPTO_PR_LAB, description="letter of intent of acquisitionfrom Transform Group"),
    DocCfg(id='EFTA01613762', author=CRYPTO_PR_LAB, description=f"WhatsApp convo with {MARIA_PRUSAKOVA}", date='2019-05-21'),
    DocCfg(id='EFTA01612721', author=CRYPTO_PR_LAB, description=f"WhatsApp convo with {MARIA_PRUSAKOVA}", date='2019-05-21'),
    EmailCfg(id='033255', description=f"{MASHA_DROKOVA}'s friend Dylan Love suggests he could quote Epstein explaining \"proper implementation of crypto might solve financial corruption issues\""),
    EmailCfg(id='EFTA01013266', description=f"{MARIA_PRUSAKOVA}'s {CRYPTO_PR_LAB} request for payment for Davos", is_interesting=True),
    EmailCfg(id='EFTA02285514', description=f"Medici Bank and {MARIA_PRUSAKOVA} meeting", is_interesting=True),
    EmailCfg(id='EFTA01013922', description=f"{CRYPTO_PR_LAB} business plan", is_interesting=True),
    EmailCfg(id='EFTA00495349', author=MARIA_PRUSAKOVA, author_reason=CRYPTO_PR_LAB, description='Ed Boyle / Medici Bank'),
    EmailCfg(
        id='EFTA02267133',
        author=LESLEY_GROFF,
        recipients=[MARIA_PRUSAKOVA],
        recipient_uncertain=f"{CRYPTO_PR_LAB} signature, could be Prusakova's partner",
        truncate_to=NO_TRUNCATE
    ),
    EmailCfg(
        id='EFTA02349697',
        author=MARIA_PRUSAKOVA,
        author_reason='boyfriend Christian, same as 032374',
        description=f'{MARIA_PRUSAKOVA} says she is starting a crypto company with {ZUBAIR_KHAN}',
    ),
    # Crypto PR Lab / Medici Bank
    DocCfg(id='EFTA01613731', description='Medici Bank conversation on Whatsapp'),
    DocCfg(id='EFTA00805569', description='Mercantile Global Holdings, San Juan Mercantile Bank & Trust, Noble Bank + Signature Bank + BitGo'),
    DocCfg(id='EFTA01088079', description="someone's thoughts on bitcoin"),
    EmailCfg(id='EFTA00495372', description=f'discussion of Medici Bank, a new successor to Noble Bank as a crypto bank', is_interesting=True),
    EmailCfg(
        id='EFTA01030070',
        description="Bo Collins inspired by Epstein to set up Mercantile Global Holdings crypto trading offering stablecoins to people who can't get bank accounts",
        is_interesting=True,
    ),
    EmailCfg(
        id='EFTA02603577',
        description='meeting with Mercantile Global Holdings Bo Collins, who would testify in the trial of Miles Guo',
        is_interesting=True
    ),
    # David Stern
    EmailCfg(
        id='EFTA02494305',
        description=f"planning a crypto exchange in Belarus, {DAVID_STERN}'s \"Chinese guys\" can't do it because of the PRC's crackdown on crypto",
    ),
    EmailCfg(id='EFTA00886763', description=f"documents are being prepared to license Epstein and {DAVID_STERN}'s crypto exchange in Belarus"),
    EmailCfg(
        id='EFTA02494000',
        description=f'Epstein says a crypto exchange in Belarus could be "very useful", {DAVID_STERN}\'s Chinese guys could have brought "significant volume"',
        is_interesting=True,
    ),
    # Honeycomb
    DocCfg(id='EFTA00803491', author=HONEYCOMB_FUND, description="deck", is_interesting=True),
    DocCfg(id='EFTA00803459', author=HONEYCOMB_FUND, description="January 2019 report", is_interesting=True),
    DocCfg(id='EFTA00603445', author=HONEYCOMB_FUND, description="July 2017 report", is_interesting=True),
    DocCfg(id='EFTA00803464', author=HONEYCOMB_FUND, description="July 2018 report", date='2018-07-17', is_interesting=True),
    # Howard Lutnick
    DocCfg(id='EFTA01582043', description=f"{DEUTSCHE_BANK} / {CANTOR} transaction"),
    DocCfg(id='EFTA01230639', date='2012-12-22', date_uncertain=True, show_full_panel=True),
    DocCfg(id='EFTA00434306', description='calendar reminder for call with Howard Lutnick', date='2011-04-26', show_full_panel=True),
    DocCfg(
        id='EFTA00020515',
        author=FBI,
        description='tip about Howard Lutnick and financial irregularities',
        is_interesting=True,
        show_full_panel=True,
    ),
    DocCfg(
        id='EFTA00289560',
        date='2012-12-28',
        description=f'{HOWARD_LUTNICK} and Epstein business deal for Adfin Solutions, Inc. (Southern Trust is Epstein)',
        truncate_to=(160_000, 162_000),
    ),
    DocCfg(
        id='EFTA01249210',
        author=FBI,
        description="tip linking American commerce secretary and Tether asset custodian Howard Lutnick to Ponzi schemes and Russian money",
        is_interesting=True,
        show_with_name=BROCK_PIERCE,
    ),
    DocCfg(
        id='EFTA01733746',
        date='2011-05-18',
        description=f'schedule showing "drinks" with {HOWARD_LUTNICK}',
        is_valid_for_name_scan=False,
        show_full_panel=True,
        truncate_to=700,
    ),
    DocCfg(
        id='EFTA01249207',
        author=FBI,
        description="tip linking American commerce secretary and Tether asset custodian Howard Lutnick to Ponzi schemes and Russian money",
        show_full_panel=True,
        show_with_name=BROCK_PIERCE,
    ),
    DocCfg(
        id='EFTA01684300',
        author=FBI,
        description=f'evidence list w/tip about {HOWARD_LUTNICK}, Cantor Fitzgerald, and BGC from an employee who had already blown the whistle successfully',
        truncate_to=(77_000, 77_300),
    ),
    EmailCfg(
        id='EFTA00173881',
        author='US Secret Service',
        recipients=['Enterprise Vetting Center'],
        show_with_name=HOWARD_LUTNICK,
        truncate_to=800
    ),
    EmailCfg(
        id='EFTA00873541',
        description=f"Epstein says his neighbor {HOWARD_LUTNICK} (recently seen with {PRINCE_ANDREW}) is smart",
        show_with_name=HOWARD_LUTNICK,
    ),
    EmailCfg(id='EFTA02374900', description=f'article about {HOWARD_LUTNICK} real estate purchase'),
    EmailCfg(id='EFTA00685520', description=f'call with {HOWARD_LUTNICK}'),
    EmailCfg(id='EFTA00436468', description=f'call with {HOWARD_LUTNICK} after he goes overseas', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA00462693', description=f"Epstein gets a table for 10 at the dinner honouring {HOWARD_LUTNICK}"),
    EmailCfg(id='EFTA00398744', description=f"{HOWARD_LUTNICK} thanks Epstein for the island vacay"),
    EmailCfg(id='EFTA00474452', description=f"{HOWARD_LUTNICK} and {JEFFREY_EPSTEIN}, NIMBYs"),
    EmailCfg(id='EFTA00398719', description=f'{HOWARD_LUTNICK} visit to the island', truncate_to=2500),
    EmailCfg(id='EFTA00401065', description=f'{HOWARD_LUTNICK} visit to the island', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA00398853', description=f'{HOWARD_LUTNICK} visit to the island', truncate_to=NO_TRUNCATE),
    EmailCfg(id='EFTA00873540', description=f'{HOWARD_LUTNICK} and {PRINCE_ANDREW}'),
    EmailCfg(id='EFTA00443752', description=f"invitation from {HOWARD_LUTNICK}"),
    EmailCfg(id='EFTA00957552', description=f'thank you note for a $50,000 donation in honor of Tether banker {HOWARD_LUTNICK}'),
    EmailCfg(id='EFTA00970606', show_with_name=HOWARD_LUTNICK),
    EmailCfg(id='EFTA01778423', show_with_name=HOWARD_LUTNICK),
    # Jem Bendell
    EmailCfg(
        id='EFTA01005603',
        author=JEM_BENDELL,
        author_reason='sig',
        highlight_quote="if you are interested in blockchain and alternative currencies",
    ),
    EmailCfg(id='EFTA00630785', description='alternative currencies discussion'),
    EmailCfg(id='EFTA02724230', description='alternative currencies discussion', truncate_to=NO_TRUNCATE),
    # Jeremy Rubin
    EmailCfg(id='EFTA02466979', description='"AID Coin"'),
    EmailCfg(id='EFTA02343669', description='rumours that people are using bitcoin to exfiltrate capital from China'),
    # Joi Ito / Kyara
    DocCfg(id='024256', author=JOI_ITO, description=f"Internet & Society: The Technologies and Politics of Control"),
    EmailCfg(id='025598', is_interesting=False),
    EmailCfg(id='028847', description=f"the announcement of a hunt for Putin's money seems to greatly concern Epstein and {JOI_ITO}"),
    EmailCfg(id='026360', description=f"{JOI_ITO} writes to Epstein to thank him for providing funding for the bitcoin core development team"),
    EmailCfg(id='025603', description=f"Epstein funnels money from {ANDREW_FARKAS} to {MIT_MEDIA_LAB} / {JOI_ITO}", truncate_to=850),
    DocCfg(id='EFTA00805860', description=f"{JOI_ITO}'s Neoteny 3 fund investor update / portfolio"),
    DocCfg(id='EFTA01118268', description=f"{JOI_ITO}'s Neoteny 3 fund investor update", date='2015-06-30'),
    DocCfg(id='EFTA01116280', description=f"{KYARA_FUND} I LLC (Physical Graph Corporation) operating agreement with edits"),
    DocCfg(id='EFTA01108775', description=f"{KYARA_FUND} I LLC (Physical Graph Corporation) operating agreement"),
    DocCfg(id='EFTA00605699', description=f"{KYARA_FUND} II LLC incorporation papers"),
    DocCfg(id='EFTA01186455', description=f"{KYARA_FUND} II LLC operating agreement"),
    DocCfg(id='EFTA00584696', description=f"{KYARA_FUND} III LLC operating agreement"),
    EmailCfg(
        id='EFTA02376268',
        description=f"introduction to {JEREMY_RUBIN}",
        recipients=[JEFFREY_EPSTEIN, JEREMY_RUBIN, LINDA_STONE, None],
    ),
    EmailCfg(id='EFTA02342561', description=f"China, {REID_HOFFMAN}, bitcoin", is_interesting=True),
    EmailCfg(id='EFTA02601259', description=f"discussion of investments hidden through {JOI_ITO}"),
    EmailCfg(id='EFTA00989593', description=f'donations from {LEON_BLACK} to {MIT_MEDIA_LAB} / Digital Currency Initiative {QUESTION_MARKS}'),
    EmailCfg(id='EFTA00999549', description=f"{JOI_ITO} and {JEREMY_RUBIN} meet {LARRY_SUMMERS} to discuss bitcoin", is_interesting=True),
    EmailCfg(id='EFTA01752601', description=f"{JOI_ITO} and Epstein name their new fund Kyara"),
    EmailCfg(id='EFTA02703885', author=LINDA_STONE, author_reason='sig', description=f"{LINDA_STONE} introducing Epstein to {JEREMY_RUBIN}"),
    EmailCfg(id='EFTA02505439', description=f'summary of {KYARA_INVESTMENT} shell companies with {JOI_ITO}'),
    EmailCfg(id='EFTA01973301', recipients=[LINDA_STONE], author_reason='signature'),
    EmailCfg(id='EFTA00473175', truncate_to=1000),
    # Kushner
    DocCfg(id='EFTA00128987', description='suspicious activity report (SAR) about Kushner co. crypto payments to suspicious Russian person'),
    # Masha Drokova
    EmailCfg(
        id='EFTA01003115',
        author=MASHA_DROKOVA,
        author_reason='"masha" in quoted',
        description='"Dylan" is probably Dylan Love who wrote a story about Epstein and bitcoin arranged by Masha Drokova',
        truncate_to=NO_TRUNCATE,
    ),
    EmailCfg(id='EFTA01010171'),
    EmailCfg(id='EFTA01010128'),
    # Misc
    DocCfg(
        id='EFTA02005667',
        author='Clearstone',
        date='2011-02-01',
        date_uncertain=True,
        description='Online Gaming Fund pitch deck',
        non_participants=[AUSTIN_HILL],
    ),
    DocCfg(
        id='EFTA01388197',
        attached_to_email_id='EFTA01388354',
        author=DEUTSCHE_BANK,
        date='2018-04-03',
        description='bitcoin and blockchain symposium',
    ),
    DocCfg(
        id='EFTA01074293',
        author='Jerry Brito, Houman Shadab, & Andrea Castillo',
        date='2014-04-10',
        description='discussion draft of "Bitcoin Financial Regulation" paper',
    ),
    DocCfg(
        id='EFTA00797937',
        author='John Pfeffer',
        date='2017-12-24',
        description="An (Institutional) Investor's Take on Cryptoassets",
    ),
    DocCfg(
        id='EFTA01612888',
        author='Kevin Peterson, Bammohan Deeduvanu, Pradip Kanjamala, and Kelly Boles',
        description='A Blockchain-Based Approach to Health Information Exchange Networks',
    ),
    DocCfg(
        id='EFTA02725909',
        date='2016-01-01',
        description=f'memo to NYDFS of NYC Bitcoin Exchange, Balaji Srinivisan & {ANDREW_FARKAS} on board',
        date_uncertain='mentions ethereum so must be after 2015-06',
    ),
    DocCfg(id='EFTA00811666', description='asset valuations of Epstein\'s holdings, includes "Coinbase via grat"', date='2018-01-31'),
    DocCfg(id='EFTA01734786', description='LedgerX Series B pitch deck'),
    DocCfg(id='EFTA00605996', description='Wedbush BUY rating on Digital Currency Group GBTC', is_interesting=False),
    EmailCfg(id='EFTA00867030', author='Kathy', description='discussion of a crypto token based on GDAX (whatever that means)', is_interesting=True),
    EmailCfg(id='EFTA01060993', description=f"Epstein says he knows \"a few\" guys who hold over $50 million in bitcoin", is_interesting=True),
    EmailCfg(id='EFTA01915234', description='discussion of crypto regulations', is_interesting=True),
    EmailCfg(id='EFTA01004753', description=f"founder of Layer1 is the only guy in crypto who doesn't want to meet Epstein (didn't last though)"),
    EmailCfg(id='EFTA02601503', description=f"discussion of investing in Layer1, an important crypto business"),
    EmailCfg(id='EFTA02517623', description=f"{VINCENZO_IOZZO} explains to Epstein how to deanonymize bitcoin", is_interesting=True),
    EmailCfg(id='EFTA00699275', description=f"'Fred' is probably Coinbase co-founder Fred Ehrsam"),
    EmailCfg(id='EFTA00830911', description='fundraising email for LedgerX which was later acquired by FTX for $298 million'),
    EmailCfg(id='EFTA01750652', description=f"discussion of using blockchains in the repo markets", is_interesting=True),
    EmailCfg(id='EFTA02002675', description=f"{JASON_CALACANIS} passes Epstein the names of the bitcoin core developers"),
    # Reid Hoffman
    EmailCfg(id='EFTA00820371', description='bank digital currencies article', is_interesting=True),
    # Russia
    EmailCfg(id='EFTA02591998', description='Epstein spoke to Vladimir Putin about digital currency', is_interesting=True),
    # SEC / NYDFS
    EmailCfg(id='EFTA00668932', description=f"pressing Farkas for a meeting with {BEN_LAWSKY_NYDFS}"),
    EmailCfg(id='EFTA01747752', description=f"Farkas delivering {BEN_LAWSKY_NYDFS}", truncate_to=700),
    EmailCfg(id='EFTA02588398', description=f"Epstein went to the Treasury Dept of the US to talk about bitcoin", is_interesting=True),
    EmailCfg(id='EFTA00987194', description=f"Epstein proposes {BEN_LAWSKY_NYDFS} get rid of sales tax on bitcoin", is_interesting=True),
    EmailCfg(id='EFTA02591315', description=f"{ANDREW_FARKAS} arranging for Epstein to meet {BEN_LAWSKY_NYDFS}", is_interesting=True, truncate_to=1900),
    EmailCfg(id='EFTA01002518', description=f"confirmation that Epstein did indeed meet {BEN_LAWSKY_NYDFS} to discuss a 'high yield product'"),
    EmailCfg(id='EFTA02587320', author=KATHRYN_RUEMMLER, description=f"Kathy Ruemmler, crypto bro-ette", is_interesting=True),
    EmailCfg(
        id='EFTA00357845',
        description=f"Epstein and {KATHRYN_RUEMMLER} meet with {BEN_LAWSKY_NYDFS} to discuss bitcoin",
        is_interesting=True,
        recipients=[LESLEY_GROFF],
        recipient_uncertain=True,
    ),
    # Sharia coin
    EmailCfg(id='032359', description='HBJ brings up "e-currency" (Sharia Coin, probably)'),
    EmailCfg(id='EFTA00964459', description=f'discussion of sharia compliant crypto token', is_interesting=True),
    EmailCfg(id='EFTA00990442', description=f"Epstein offers to cover all costs for Sharia Coin", is_interesting=True),
    EmailCfg(id='EFTA02396341', description='the Sharia Coin gambit', is_interesting=True),
    EmailCfg(id='EFTA01005117', description='announcement that the Stellar blockchain is Sharia compliant'),
    # Steven Sinofsky / a16z
    EmailCfg(id='EFTA01804670', highlight_quote='Andreessen really into this whole virtual money space', truncate_to=700),
    # Valar
    DocCfg(
        id='EFTA00591045',
        attached_to_email_id='EFTA01001339',
        author=VALAR_FUND,
        description='pitch deck',
        non_participants=[MASAYOSHI_SON],
    ),
    DocCfg(id='EFTA00810239', author=VALAR_FUND, description='pitch deck'),
    DocCfg(id='EFTA00810510', author=VALAR_FUND, description='Fall 2016 Update'),
    DocCfg(id='EFTA00810474', author=VALAR_FUND, description='Fall 2018 Update'),
    DocCfg(id='EFTA01121910', author=VALAR_VENTURES, description="contract", truncate_to=DEFAULT_TRUNCATE_TO),
    DocCfg(id='EFTA00808277', author=VALAR_VENTURES, description="contract", truncate_to=DEFAULT_TRUNCATE_TO),
    DocCfg(id='EFTA01088484', author=VALAR_VENTURES, description="contract", truncate_to=DEFAULT_TRUNCATE_TO),
    DocCfg(id='EFTA00591691', author=VALAR_VENTURES, description="contract", truncate_to=DEFAULT_TRUNCATE_TO),
    DocCfg(id='EFTA00810362', author=VALAR_VENTURES, description="investor questionnaire", truncate_to=DEFAULT_TRUNCATE_TO),
    EmailCfg(id='EFTA01001339', description=f"{PETER_THIEL} introduces Epstein to the {VALAR_VENTURES} founders"),
    # Vincenzo
    EmailCfg(id='EFTA00637023', description=f"discussion of getting around money laundering laws in places like Myanmar + Mongolia"),
    EmailCfg(id='EFTA02588723', description=f'discussion of crypto food stamps debit cards', is_interesting=True),
    EmailCfg(id='EFTA02588748', description=f'discussion of crypto food stamps debit cards', is_interesting=True, truncate_to=800),
    EmailCfg(id='EFTA02584771', description=f"discussion of decentralized prediction markets (e.g. Polymarket)", is_interesting=True),
    EmailCfg(id='EFTA01008242', description='Epstein passes on an investment in Radius'),
    EmailCfg(id='EFTA00995269', description=f"Epstein suggests Monaco and Vatican City as good places for crypto"),
    # ZCash / Madars Virza
    DocCfg(id='EFTA00811130', author=PERKINS_COIE, description='tax opinion on ZCash tokens'),
    DocCfg(id='EFTA00603348', description=f"Electric Coin Company created the untraceable crypto ZCash funded by {LARRY_SUMMERS}'s DCG"),
    DocCfg(id='EFTA00803405', description=f"Honeycomb Asset Management fund brochure", truncate_to=DEFAULT_TRUNCATE_TO),
    EmailCfg(id='EFTA02282357', author=DAPHNE_WALLACE, author_uncertain='could be Lesley'),
    EmailCfg(id='EFTA02283028', author=DAPHNE_WALLACE, author_uncertain='could be Lesley'),
    EmailCfg(id='EFTA02620980', description=f"{MADARS_VIRZA} says Tether was created to avoid anti-money laundering laws"),
    EmailCfg(id='EFTA01781741', description=f'{JOI_ITO} introduces Epstein to {MADARS_VIRZA} and the untraceable private crypto ZCash'),
    EmailCfg(id='EFTA02645742', description=f"tax implications of the untraceable 'privacy coin' ZCash", is_interesting=True),
    EmailCfg(
        id='EFTA02374960',
        description=f"Epstein says that he will fund a seat at MIT Media Lab for ZCash founder {MADARS_VIRZA}",
        is_interesting=True,
    ),
    # Unsorted
    EmailCfg(
        id='EFTA00096009',
        author='SDNY Cybercrimes',
        description='Epstein legal defense team requested bail hearing transcript of OneCoin fraudster Konstantin Ignatova',
        recipients=[USANYS],
    ),
    EmailCfg(id='EFTA00629471', author=JOHN_BROCKMAN, non_participants=[AUSTIN_HILL, 'David Geffen'], truncate_to=3000),
    EmailCfg(id='EFTA01762201', highlight_quote='indoctrinating kids into an economy', show_with_name=BOBBY_KOTICK),
    EmailCfg(id='EFTA02414991', description=f"Epstein signs up for World of Warcraft", truncate_to=350),
    EmailCfg(id='EFTA01388354', description=f"invitation to {DEUTSCHE_BANK} blockchain event with Mike Novogratz"),
    EmailCfg(id='EFTA01434500', description=f"everyone assumes Epstein's banker {PAUL_BARRETT} will want to know about the blockchain event"),
    EmailCfg(id='EFTA00993615', description=f"{MASHA_DROKOVA} explains the price of bitcoin can be manipulated if Epstein makes public comments"),
    EmailCfg(id='EFTA01784901', description=f"{JEREMY_RUBIN} has cashed multiple checks directly from Epstein", is_interesting=True),
    EmailCfg(
        id='EFTA01007544',
        description=f'{JEREMY_RUBIN} describes "grey area between pump and develop" when Epstein objects on ethical grounds',
        is_interesting=True,
    ),
    EmailCfg(
        id='EFTA01748990',
        description=f"{AUSTIN_HILL} is upset that Ito and Epstein seem to have invested in Stellar and Ripple, competing blockchain ecosystems",
        is_interesting=True,
    ),
    EmailCfg(id='EFTA01943067', description=f'comments on {ARIANE_DE_ROTHSCHILD} and bitcoin', is_interesting=True),
    EmailCfg(id='EFTA00835324', recipients=[SERGEY_BELYAKOV], description='assessment of Mycelium bitcoin wallet'),
    EmailCfg(id='EFTA00368659', description=f"meeting with {REID_HOFFMAN} about bitcoin", is_interesting=True),
    EmailCfg(id='EFTA01039626'),
    EmailCfg(
        id='EFTA02035756',
        description=f"{AL_SECKEL} introduces Epstein to Bannon business partner {BROCK_PIERCE} of Tether/{BLOCKCHAIN_CAPITAL}",
        is_interesting=True,
    ),
    EmailCfg(id='EFTA02372964', description='an "ICO" or "initial coin offering" was a very popular type of crypto scam at this time'),
    EmailCfg(
        id='EFTA02177147',
        author=LINDA_STONE,
        author_uncertain='"Hi Linda"',
        description=f"Epstein planning a symposium \"on Alternative Money or Complementary Currency\" in 2012",
        is_interesting=True,
        recipients=[LESLEY_GROFF],
    ),
    EmailCfg(
        id='EFTA01965506',
        author=LINDA_STONE,
        author_reason='"iPhone word substitution" in signature, which is traced back to "Linda, thanks" in EFTA00961792',
        description='Barry Silbert became the CEO of Digital Currency Group, a huge US crypto fund w/Larry Summers on the board',
        is_interesting=True,
    ),
    EmailCfg(id='EFTA00899331', truncate_to=NO_TRUNCATE),
]

# Chris Dilorio emails
WHISTLEBLOWER_IDS = {
    'EFTA00093702_37': '',
    'EFTA00093702_13': '',
    'EFTA00093702_21': RIOT_BLOCKCHAIN_DESCRIPTION,
    'EFTA00093702_30': f"Signature Bank failed in 2023, {RIOT_BLOCKCHAIN_DESCRIPTION}",
    'EFTA00093702_29': RIOT_BLOCKCHAIN_DESCRIPTION,
    'EFTA00093702_28': RIOT_BLOCKCHAIN_DESCRIPTION,
    'EFTA00093702_27': RIOT_BLOCKCHAIN_DESCRIPTION,
    'EFTA00093702_26': RIOT_BLOCKCHAIN_DESCRIPTION,
}


CRYPTO_CFGS.extend([whistleblower_cfg(id, desc) for id, desc in WHISTLEBLOWER_IDS.items()])



"""
Mobile vs. non-mobile.
"""
from dataclasses import dataclass
from datetime import datetime
from typing import ClassVar

from rich.padding import PaddingDimensions
from rich.text import Text

from epstein_files.util.constant.strings import SUBHEADER_STYLE, TIMESTAMP_STYLE
from epstein_files.util.helpers.link_helper import link_markup
from epstein_files.util.helpers.string_helper import capitalize_first, iso_timestamp, starred_header

ALL_OTHER_FILES_MULTIPLIER = 1.5  # applied when --all-other-files is in use
DEFAULT_WIDTH = 150

RICH_LINK_MARKUP = link_markup('https://pypi.org/project/rich/', 'rich')
MOBILE_WARNING = f"This page was generated by {RICH_LINK_MARKUP}. It is very poorly optimized for mobile and may take a while to render on a phone." \
                  "\nIf your browser won't render part of the page try reloading."
NOT_ALL_FILES_MSG = 'Not All Epstein Files Are Here!'


# TODO: another thing that changes based on mobile is id_only for link texts
@dataclass
class MobileConfig:
    """
    show_with_indent (int): Indent for `show_with_name` files
    """
    abbreviations_width: ClassVar[int | None] = None
    attachment_indent: ClassVar[int] = 6
    character_bio_padding: ClassVar[tuple[int, int, int, int]] = (0, 0, 1, 2)
    email_info_in_subtitle: ClassVar[bool] = False
    info_indent: ClassVar[int] = 1
    max_alt_links: ClassVar[int | None] = 1
    not_all_files_warning: ClassVar[str] = starred_header(NOT_ALL_FILES_MSG, num_spaces=1, num_stars=1)
    num_color_key_cols: ClassVar[int] = 2
    other_files_preview_chars: ClassVar[int] = 300
    other_files_table_indent: ClassVar[int] = 0
    section_header_padding: ClassVar[PaddingDimensions] = (2, 1, 1, 1)
    show_with_indent: ClassVar[int] = 0
    show_emailer_tables: ClassVar[bool] = False
    site_glossary_horizontal_padding: ClassVar[int] = 2
    social_link_separator: ClassVar[str] = ' '
    subtitle_width: ClassVar[int | None] = None
    suppressed_file_indent: ClassVar[int] = 0
    width: ClassVar[int] = 52

    @classmethod
    def format_text_msg_time(cls, dt: datetime) -> str:
        return dt.strftime(r"%-d/%-m/%-y %-H:%M")

    @classmethod
    def info_padding(cls) -> tuple[int, int, int, int]:
        return (0, 0, 0, cls.info_indent)

    @classmethod
    def email_subheader(cls, email_type: str, author: Text, recipients: Text, timestamp: datetime | Text) -> Text:
        prefix = f"{capitalize_first(email_type)} from " if email_type != 'email' else ''
        txt = Text(prefix, SUBHEADER_STYLE).append(author)
        timestamp = timestamp if isinstance(timestamp, Text) else Text(' at ').append(str(timestamp), TIMESTAMP_STYLE)
        txt.append(' to ').append(recipients).append(timestamp)
        return txt

    @classmethod
    def suppressed_file_padding(cls) -> tuple[int, int, int, int]:
        return (0, 0, 0, cls.suppressed_file_indent)


class SiteConfig(MobileConfig):
    abbreviations_width: ClassVar[int | None] = 62
    attachment_indent: ClassVar[int] = 12
    character_bio_padding: ClassVar[tuple[int, int, int, int]] = (0, 1, 0, 0)
    email_info_in_subtitle: ClassVar[bool] = True
    info_indent: ClassVar[int] = 1
    max_alt_links: ClassVar[int | None] = None
    not_all_files_warning: ClassVar[str] = starred_header(NOT_ALL_FILES_MSG, num_spaces=6, num_stars=14)
    num_color_key_cols: ClassVar[int] = 6
    other_files_preview_chars: ClassVar[int] = 900
    other_files_table_indent: ClassVar[int] = 2
    section_header_padding: ClassVar[PaddingDimensions] = (2, 25, 1, 25)
    show_emailer_tables: ClassVar[bool] = True
    show_with_indent: ClassVar[int] = 30
    site_glossary_horizontal_padding: ClassVar[int] = 5
    social_link_separator: ClassVar[str] = '  /  '
    subtitle_width: ClassVar[int | None] = 110
    suppressed_file_indent: ClassVar[int] = 0
    width: ClassVar[int] = DEFAULT_WIDTH

    @classmethod
    def format_text_msg_time(cls, dt: datetime) -> str:
        return iso_timestamp(dt)

    @classmethod
    def email_subheader(cls, email_type: str, author: Text, recipients: Text, timestamp: datetime) -> Text:
        return super().email_subheader(
            f"OCR text of {email_type}",
            author,
            recipients,
            Text(f" probably sent at ").append(str(timestamp), style=TIMESTAMP_STYLE)
        )



"""
Classes holding information that creates color highlighting via `rich.Highlighter`
regex mechanism as well as identify individuals in email headers etc.
"""
import json
import re
from abc import ABC
from dataclasses import dataclass, field

from epstein_files.people.contact import Contact
from epstein_files.people.names import Name, constantize_name
from epstein_files.util.constant.strings import REGEX_STYLE_PREFIX
from epstein_files.util.helpers.data_helpers import without_falsey
from epstein_files.util.helpers.string_helper import as_pattern, capture_group_marker
from epstein_files.util.logging import logger


@dataclass(kw_only=True)
class HighlightGroup(ABC):
    """
    Regex and style information for things we want to highlight.

    Attributes:
        label (str): RegexHighlighter match group name
        regex (re.Pattern): regex pattern identifying strings matching this group
        style (str): Rich style to apply to text matching this group
        theme_style_name (str): The style name that must be a part of the rich.Console's theme
    """
    label: str = ''
    regex: re.Pattern = field(init=False)
    style: str
    theme_style_name: str = field(init=False)
    _capture_group_label: str = field(init=False)
    _capture_group_marker: str = field(init=False)

    def __post_init__(self):
        if not self.label:
            raise ValueError(f'Missing label for {self}')

        self._capture_group_label = self.label.lower().replace(' ', '_').replace('-', '_')
        self._capture_group_marker = capture_group_marker(self._capture_group_label)
        self.theme_style_name = f"{REGEX_STYLE_PREFIX}.{self._capture_group_label}"


@dataclass(kw_only=True)
class HighlightPatterns(HighlightGroup):
    """
    Color highlighting for things other than people's names (e.g. phone numbers, email headers).

    Attributes:
        patterns (list[str]): regex patterns identifying strings matching this group
        regex_flags (re.RegexFlag): flags to use when compiling the patterns to an `re.Pattern`
        use_word_boundary (bool, optional): if True, patterns can only match before/after word boundary `\b`
    """
    patterns: list[str] = field(default_factory=list)
    regex_flags: re.RegexFlag = re.IGNORECASE | re.MULTILINE
    use_word_boundary: bool = False
    _pattern: str = field(init=False)

    def __post_init__(self):
        super().__post_init__()

        if not self.label:
            raise ValueError(f"No label provided for {repr(self)}")

        self.patterns = [as_pattern(p) for p in self.patterns]
        self._pattern = '|'.join(self.patterns)

        if self.use_word_boundary:
            self._pattern = fr"\b(({self._pattern})s?)\b"

        self.regex = self.compile_patterns(self._pattern)


    def __repr__(self) -> str:
        return f"{type(self).__name__}(label='{self.label}', pattern='{self._pattern}', style='{self.style}')"

    def compile_patterns(self, pattern: str) -> re.Pattern:
        try:
            return re.compile(fr"({self._capture_group_marker}{pattern})", self.regex_flags)
        except Exception as e:
            logger.error(f"Failed to compile regex '{pattern}'\n\nTrying each piece individually...")

            for p in self.patterns:
                logger.warning(f"Attempting to compile '{p}'...")
                re.compile(p)

            raise e


@dataclass(kw_only=True)
class HighlightedNames(HighlightPatterns):
    """
    Encapsulates info about people, places, and other strings we want to highlight with RegexHighlighter.
    Constructor must be called with either an 'emailers' arg or a 'pattern' arg (or both).

    Attributes:
        category (str): optional string to use as an override for self.label in some contexts
        contacts (list[ContactInfo]): optional `ContactInfo` objects with names and regexes
        contacts_lookup (dict[Name, ContactInfo]): lookup dictionary for `ContactInfo` objects
        should_match_first_last_name (bool): if False don't match first/last/reversed versions of emailers
    """
    category: str = ''
    contacts: list[Contact] = field(default_factory=list)
    contacts_lookup: dict[Name, Contact] = field(default_factory=dict)
    flags: re.RegexFlag = re.IGNORECASE
    should_match_first_last_name: bool = True  # TODO: this no longer does anything?

    def __post_init__(self):
        if not (self.patterns or self.contacts):
            raise ValueError(f"Must provide either 'contacts' or 'patterns' arg.")
        elif not self.label:
            if len(self.contacts) == 1 and self.contacts[0].name:
                self.label = self.contacts[0].name
            else:
                raise ValueError(f"No label provided for {repr(self)}")

        super().__post_init__()
        with_contacts_pattern = '|'.join([c.highlight_pattern for c in self.contacts] + self.patterns)
        self._pattern = fr"\b(({with_contacts_pattern})s?)\b"
        self.regex = self.compile_patterns(self._pattern)
        self.contacts_lookup = Contact.build_name_lookup(self.contacts)

        for contact in self.contacts:
            contact.category = self.category_str
            contact.style = self.style

        logger.debug(repr(self))

    @property
    def category_str(self) -> str:
        if self.category:
            return self.category
        elif len(self.contacts) == 1 and self.label == self.contacts[0].name:
            return ''
        else:
            return self.label.replace('_', ' ')

    def info_for(self, name: str, include_category: bool = False) -> str | None:
        """Label and additional info for 'name' if 'name' is in `self.contacts`."""
        info_pieces = [self.category_str] if include_category else []

        if (contact := self.contacts_lookup.get(name)):
            # Don't prefix with category if category is in the info string
            if info_pieces and info_pieces[0] in contact.info:
                info_pieces = [contact.info]
            else:
                info_pieces.append(contact.info)

        info_pieces = without_falsey(info_pieces)
        return ', '.join(info_pieces) if info_pieces else None

    def __repr__(self) -> str:
        s = f"{type(self).__name__}("

        for property in ['label', 'style', 'category', 'patterns', 'contacts', '_pattern']:
            value = getattr(self, property)

            if not value or (property == 'label' and len(self.contacts) == 1 and not self.patterns):
                continue

            s += f"\n    {property}="

            if isinstance(value, dict):
                s += '{'

                for k, v in value.items():
                    s += f"\n        {constantize_name(k)}: {json.dumps(v).replace('null', 'None')},"

                s += '\n    },'
            elif property == 'patterns':
                s += '[\n        '
                s += repr(value).removeprefix('[').removesuffix(']').replace(', ', ',\n        ')
                s += ',\n    ],'
            elif isinstance(value, list) and value and isinstance(value[0], Contact):
                s += '[\n        '
                s += f"    {', '.join([c.name for c in value])}"
                s += ',\n    ],'
            else:
                s += f"{json.dumps(value)},"

        return s + '\n)'

    def __str__(self) -> str:
        return super().__str__()


@dataclass(kw_only=True)
class ManualHighlight(HighlightGroup):
    """For when you can't construct the regex."""
    pattern: str
    regex_flags: re.RegexFlag = re.MULTILINE

    def __post_init__(self):
        super().__post_init__()

        if self._capture_group_marker not in self.pattern:
            raise ValueError(f"Label '{self.label}' must appear in regex pattern '{self.pattern}'")

        self.regex = re.compile(self.pattern, self.regex_flags)



from copy import deepcopy
from pathlib import Path
from typing import Mapping

from rich.console import RenderableType
from rich.style import Style
from rich.table import Table
from rich.text import Text

from epstein_files.output.html.elements import *
from epstein_files.util.constant.html import CUSTOM_HTML_TEMPLATE, HTML_TERMINAL_THEME
from epstein_files.util.helpers.data_helpers import listify
from epstein_files.util.helpers.file_helper import log_file_write
from epstein_files.util.helpers.string_helper import join_truthy
from epstein_files.util.logging import logger

CSS = Path(__file__).parent.joinpath('page.css').read_text()
BOX_BORDER_RADIUS = 4
BOX_BORDER_WIDTH = 2
BORDER_HORIZONTAL_PADDING = to_em(1)
BORDER_VERTICAL_PADDING = to_em(0.5)
VERTICAL_MARGIN = to_em(1.9)  # Between elements
BOTTOM_MARGIN_PROPS = {'margin-bottom': VERTICAL_MARGIN}
VERTICAL_MARGIN_PROPS = vertical_margin_props(VERTICAL_MARGIN)

PANEL_BASE_PROPS = {
    "display": "inline-block",
    'width': 'fit-content',
    **padding_props(BORDER_HORIZONTAL_PADDING, BORDER_VERTICAL_PADDING)
}

PANEL_BORDER_PROPS = {
    "border-radius": to_px(BOX_BORDER_RADIUS),
    "border-style": "solid",
    "border-width": to_px(BOX_BORDER_WIDTH),
}


def body_panel_css_props(border_style: str | Style | None) -> dict[str, str]:
    """CSS for panel with Document text."""
    return {
        **border_css_props(border_style),
        **PANEL_BASE_PROPS,
    }


def border_css_props(style: str | Style | None) -> dict[str, str]:
    if style is None:
        return {}

    html_style = HtmlStyle(style)

    if html_style.hex:
        return {
            "border-color": html_style.hex,
            **PANEL_BORDER_PROPS
        }
    else:
        return {}


def buffer_as_html(_console: Console, clear: bool = True) -> str:
    html = _console.export_html(
        clear=clear,
        code_format=PRE_CONSOLE_TEMPLATE,
        inline_styles=True,
        theme=HTML_TERMINAL_THEME,
    )

    return html.split(SPLITTER)[0]  # Strip cruft we need to have templated to get export_html() to work


def rich_to_html(
    obj: RenderableType,
    props: CssProps = None,
    minimize_width: bool = False,
    maximize_width: bool = False
) -> str:
    """Convert a rich renderable to HTML string."""
    props = props or {}
    old_console_width = html_console.width
    html_console.width = from_em(props.get('max-width') or props.get('width')) or html_console.width

    if minimize_width:
        measurement = html_console.measure(obj)
        logger.debug(f"renderable has measurement: {measurement}")
        html_console.width = measurement.minimum
    elif maximize_width:
        html_console.width = 1024

    if html_console.width != old_console_width:
        logger.debug(f"Temporarily set console width to {html_console.width}")

    html_console.print(obj, end='')
    html_text = buffer_as_html(html_console)
    html_console.width = old_console_width
    return html_text


def table_to_html(table: Table, css_props: CssProps = None) -> str:
    col_styles = [col.style or '' for col in table.columns]
    border_props = border_css_props(table.border_style)
    header_css_props = HtmlStyle(table.header_style).to_css if table.header_style else {}
    row_props = {'border-bottom': '1px solid dimgray'} if table.show_lines else {}

    if table.title:
        title_txt = table.title.copy() if isinstance(table.title, Text) else Text(table.title)

        if table.title_style:
            title_txt.stylize(table.title_style)

        title_html = rich_to_html(title_txt)
    else:
        title_html = ''

    cell_props = {
        **row_props,
        **(vertical_pad_props(to_em(0.5)) if table.show_lines else {})
    }

    headers = [
        table_cell(
            Text('', style=col.header_style or '').append(col.header),
            {
                'max-width': to_em(col.max_width or col.width),
                'text-align': ((col.header.justify or '') if isinstance(col.header, Text) else '') or col.justify,
                **header_css_props
            },
            'columnheader'
        )
        for col in table.columns
    ]

    rows = [
        [
            table_cell(
                Text('', style=col_styles[j]).append(col._cells[i]) if isinstance(col._cells[i], (str, Text)) else col._cells[i],
                {
                    'max-width': to_em(col.max_width or col.width),
                    'text-align': table.columns[j].justify,
                    **cell_props
                }
            )
            for j, col in enumerate(table.columns)
        ]
        for i, _row in enumerate(table.rows)
    ]

    rows = [headers, *rows] if table.show_header else rows
    html_rows = [div_class('\n'.join(row), 'row', row_props, role='row') for row in rows]
    table_html = div_class('\n\n'.join(html_rows), 'table', border_props, role='table')
    div_props = {**BOTTOM_MARGIN_PROPS, **(css_props or {})}
    return div_class(title_html + '\n' + table_html, 'table_container', div_props)


def table_cell(contents: RenderableType, props: CssProps = None, extra_class: str = '') -> str:
    cell_html = rich_to_html(contents, props)
    return div_class(cell_html, join_truthy(extra_class, 'column'), props, role='cell')


def text_to_div(txt: Text, css_props: dict[str, str], class_name: str = f"{BLACK_BACKGROUND} no_expand") -> str:
    return div_class(rich_to_html(txt), class_name, css_props)


def one_row_table_html(table: Table, css_props: CssProps = None) -> str:
    """For printing one row/one column tables currently generated by Email objs with description."""
    border_props = border_css_props(table.border_style)
    col = table.columns[0]

    if table.show_header:
        header_props = {
            'border-bottom-color': HtmlStyle(table.border_style).hex,
            **(HtmlStyle(table.header_style).to_css if table.header_style else {})
        }

        logger.debug(f"header_props: {header_props}\nborder_props: {border_props}\n")

        header_div = div_class(
            rich_to_html(Text('', style=col.header_style or '').append(col.header)),
            'document_panel_header',
            {'text-align': col.header.justify, **{**header_props}}
        )
    else:
        header_div = ''

    body_div = div_class(
        rich_to_html(Text('', style=col.style).append(col._cells[0])),
        'no_expand document_body_container',
        PANEL_BASE_PROPS
    )

    return div_class(
        header_div + '\n' + body_div,
        f'{BLACK_BACKGROUND} no_expand',
        {
            'text-align': table.columns[0].justify,
            **border_props,
            **(css_props or {})
        }
    )


def text_to_list(elements: list[str] | list[Text], tag: str = 'ul', **kwargs) -> str:
    if elements and isinstance(elements[0], Text):
        elements = [rich_to_html(txt, maximize_width=True) for txt in elements]
        elements = [strip_outer_tag(t, 'pre') for t in elements]

    return list_tag(elements, **kwargs)


def write_templated_html(divs: list[str] | str, output_path: Path) -> Path:
    body = '\n\n'.join(listify(divs))

    html = str(CUSTOM_HTML_TEMPLATE).format(
        code=body,
        stylesheet=CSS,
        background=HTML_TERMINAL_THEME.background_color.hex,
        foreground=HTML_TERMINAL_THEME.foreground_color.hex,
    )

    output_path.write_text(html)
    log_file_write(output_path)
    return output_path



import re
from dataclasses import dataclass, field, fields
from typing import Literal, Self

from rich.text import Text

from epstein_files.people.names import Name, constantize_name, extract_first_name, extract_last_name
from epstein_files.util.constant.strings import INDENT_NEWLINE, INDENTED_JOIN, LAW_ENFORCEMENT, PartialName
from epstein_files.util.helpers.data_helpers import constantize_names
from epstein_files.util.helpers.link_helper import link_text_obj
from epstein_files.util.helpers.string_helper import as_pattern, indented, is_integer, quote, remove_question_marks, join_truthy
from epstein_files.util.logging import logger

MIN_LEN_FOR_OPTIONAL_LAST_CHAR = 5
MGMT_PATTERN = r"M(ana)?ge?m(en)?t"
MGMT_REGEX = re.compile(MGMT_PATTERN)
COMPANY_SUFFIX_REGEX = re.compile(fr".*?(,? (Inc\.?|LLC|{MGMT_PATTERN}))$")
MIDDLE_INITIAL_REGEX = re.compile(r"^[A-Z]\.?$")
SIMPLE_NAME_REGEX = re.compile(r"^[-\w, ]+$", re.IGNORECASE)


@dataclass
class Contact:
    """
    Attributes:
        name (str): Person or organization name
        info (str, optional): biographical info about this person
        emailer_pattern (str, optional): manually constructed regex pattern to match this person in email headers
        category (str, optional): category of this entity
        style (str, optional): style to use when printing this entity's name
        emailer_regex (re.Pattern): pattern that matches this person's name in email headers
        highlight_regex (re.Pattern): pattern that matches this person's name in various variations for highlightihng
        is_emailer (bool): should email headers be scanned for this entity
        is_interesting (bool): should a biographical entry be generated for this panel in the chronological view
        is_junk (bool): for junk email
        is_organization (bool): if this is a company or group, don't try to match first and last versions of its name
        link_to_bio (str, optional): a link to some info about this entity
        match_partial (PartialName | None): whether to also match this entity's first and last names
    """
    name: str
    info: str = ''
    emailer_pattern: str = ''
    category: str = ''
    style: str = ''
    emailer_regex: re.Pattern = field(init=False)
    highlight_regex: re.Pattern = field(init=False)
    is_emailer: bool = True
    is_interesting: bool = True  # Eligible for bio panel
    is_junk: bool = False  # TODO: this sucks
    is_organization: bool = False
    link_to_bio: str = ''
    match_partial: PartialName | None = 'last'
    # jmail_url: str

    def __post_init__(self):
        if self.is_organization:
            self.match_partial = None

        try:
            self.emailer_regex = re.compile(self.pattern, re.IGNORECASE)
        except re.error as e:
            logger.fatal(f"failed to compile emailer_regex for {self.name}: {e}")
            raise e

        try:
            self.highlight_regex = re.compile(fr"\b({self.highlight_pattern})\b", re.IGNORECASE)
        except re.error as e:
            logger.fatal(f"failed to compile highlight_regex for {self.name}: {e}")
            raise e

    @property
    def bio(self) -> Text:
        """Biographical info about this entity."""
        from epstein_files.output.epstein_highlighter import non_epstein_highlighter
        txt = Text('')

        if self.link_to_bio:
            txt.append(link_text_obj(self.link_to_bio, self.name, self.bold_style))
        else:
            txt.append(self.name, style=self.style)

        if self.category:
            txt.append(' [', style='dim').append(self.category.lower(), style=f'{self.style} dim').append(']', style='dim')

        return txt.append(' ').append(non_epstein_highlighter(Text(self.info, style='italic grey70')))

    @property
    def bold_style(self) -> str:
        return f"{self.style} bold".strip()

    @property
    def has_bio(self) -> bool:
        return bool(self.is_interesting and self.info)

    @property
    def highlight_pattern(self) -> str:
        """
        `self.emailer_pattern` extended with first/last name variations.
        Used for color highlighting with `HighlightedNames` / `EpsteinHighlighter`.
        """
        # TODO: this sucks
        if self.is_junk:
            return self.pattern

        return '|'.join(self._name_patterns)

    @property
    def pattern(self) -> str:
        """Pattern used for matching emails, base pattern for highlights."""
        if self.emailer_pattern:
            pattern = self.emailer_pattern
        else:
            if self._middle_initial:
                pattern = fr"{self._names[0]} ({self._middle_initial}\.? )?{self._names[-1]}"
            else:
                pattern = self.name

            if len(self.name) >= MIN_LEN_FOR_OPTIONAL_LAST_CHAR and not self.is_organization:
                pattern += '?'

        return as_pattern(pattern)

    @property
    def _middle_initial(self) -> str:
        if len(self._names) != 3:
            return ''
        elif MIDDLE_INITIAL_REGEX.match(self._names[1]):
            return self._names[1][0]
        else:
            return ''

    @property
    def _name_patterns(self) -> list[str]:
        """['Firstname', 'Lastname', 'Lastname, Firstname'."""
        name_patterns = [self.pattern]

        if ' ' in self.name and self.match_partial is not None:
            name = remove_question_marks(self.name)  # TODO: this sucks
            first_name = extract_first_name(name)
            last_name = extract_last_name(name)
            name_patterns.append(as_pattern(f"{last_name},? {first_name}"))  # Reversed name

            if self.match_partial in ['both', 'first'] and SIMPLE_NAME_REGEX.match(first_name):
                name_patterns.append(as_pattern(first_name))

            if self.match_partial in ['both', 'last'] and SIMPLE_NAME_REGEX.match(last_name):
                name_patterns.append(as_pattern(last_name))

        logger.debug(f"Contact('{self.name}') name_patterns: '{name_patterns}'")
        return name_patterns

    @property
    def _names(self) -> list[str]:
        return self.name.split()

    @property
    def _props_strs(self) -> list[str]:
        props = []

        def add_prop(f, value):
            if self.emailer_pattern:
                props.append(f"{f.name}={value}")
            else:
                props.append(value)

        for _field in fields(self):
            if _field.name == 'name':
                add_prop(_field, constantize_name(getattr(self, _field.name)))
            elif (value := getattr(self, _field.name)) and not _field.name.endswith('regex'):
                if _field.name.endswith('pattern'):
                    value = f'r"{value}"'
                else:
                    value = quote(constantize_names(str(value)))
                    value = ('f' + value) if '{' in value else value

                add_prop(_field, value)

        props.append(f'highlight_pattern=r"{self.highlight_pattern}"')
        return props

    def __repr__(self) -> str:
        props = self._props_strs
        type_str = f"{type(self).__name__}("

        if self.emailer_pattern:
            repr_str = f"{type_str}{INDENT_NEWLINE}" + INDENTED_JOIN.join(props)
            repr_str += ',' if props else ''
            repr_str += '\n)'
        else:
            repr_str = type_str + ', '.join(props) + ')'

        return repr_str

    @classmethod
    def build_name_lookup(cls, contacts: list[Self]) -> dict[Name, Self]:
        """Dict of `Contact` objects keyed by contact name."""
        return {c.name: c for c in contacts}

    @classmethod
    def repr_string(cls, contact_infos: list[Self]) -> str:
        return '[\n' + indented(',\n'.join([repr(contact) for contact in contact_infos]), 4) + '\n],'


def acronym(name: str, description: str = '', **kwargs) -> Contact:
    """Like organization() but auto-generates a regex matching the org's initials."""
    initials = [word[0] for word in name.split() if word[0].isupper()]
    initials_pattern = ''.join([fr"{letter}\.?" for letter in initials])

    return organization(
        ''.join(initials),
        join_truthy(name, description, ', '),
        '|'.join([initials_pattern, name]),
        **kwargs
    )


# TODO: make class method (?)
def organization(name: str, description: str = '', emailer_pattern: str = '', **kwargs) -> Contact:
    if (suffix_match := COMPANY_SUFFIX_REGEX.match(name)) and not emailer_pattern:
        suffix = suffix_match.group(1)
        emailer_pattern = name.removesuffix(suffix)

        if not MGMT_REGEX.search(suffix):
            if suffix.startswith(','):
                suffix = suffix.replace(',', ',?')
            else:
                suffix = f",?{suffix}"

        if suffix.endswith('.'):
            suffix = suffix.replace('.', r'\.?')

        emailer_pattern += fr"({suffix})?"

    kwargs['is_emailer'] = kwargs.get('is_emailer', False)
    return Contact(name, description, emailer_pattern, is_organization=True, **kwargs)


def epstein_co(name: str, emailer_pattern: str = '', description: str = '', **kwargs) -> Contact:
    return organization(name, join_truthy('Epstein company', description), emailer_pattern, **kwargs)


def epstein_trust(
    name: str,
    emailer_pattern: str = '',
    beneficiaries: list[str] | None = None,
    trustees: list[str] | None = None,
) -> Contact:
    """One of Epstein's financial trust entities."""
    name = f"Jeffrey E. Epstein {name} Trust" if is_integer(name) else name
    beneficiary_str = ''

    if beneficiaries:
        if len(beneficiaries) == 1:
            beneficiary_str = f"sole beneficiary {beneficiaries[0]}"
        else:
            beneficiary_str = f"beneficiaries {', '.join(beneficiaries)}"

    if trustees:
        beneficiary_str = join_truthy(beneficiary_str, f"trustees: " + ', '.join(trustees), ', ')

    beneficiary_str = f", {beneficiary_str}" if beneficiary_str else ''
    return organization(name, f'Epstein financial trust{beneficiary_str}', emailer_pattern)


def law_enforcement(name: str, emailer_pattern: str = '', description: str = '') -> Contact:
    return organization(name, description or LAW_ENFORCEMENT, emailer_pattern, is_interesting=False)



"""
Epstein money related files. This category makes is_interesting = True.
"""
from epstein_files.documents.config.doc_cfg import DocCfg, EmailCfg
from epstein_files.people.names import *
from epstein_files.util.constant.strings import *


MONEY_CFGS = [
    # private placement memoranda
    DocCfg(
        id='024432',
        date='2006-09-27',
        description=f"Michael Milken's Knowledge Universe Education (KUE) $1,000,000 corporate share placement notice (SEC filing?)"
    ),
    DocCfg(id='024003', description=f"New Leaf Ventures ($375 million biotech fund) private placement memorandum"),

    # DOJ files
    DocCfg(
        id='EFTA01413294',
        comment='related to EFTA01357341, efta01363125, + more based on Vavilov Street address',
        date='2016-12-16',
        date_uncertain='hard to tell',
        description='compliance report on payment from Epstein to Sberbank account of Nikolay Aleksandrovich Gyrov',
    ),
    DocCfg(
        id='EFTA00173953',
        author='Organized Crime Drug Enforcement Task Force',
        description='report on DEA investigations into Epstein related drug money laundering',
    ),
    DocCfg(
        id='EFTA01681865',
        author=DEUTSCHE_BANK,
        description="explanations of all of Epstein's large payments prepared for DOJ",
        is_interesting=True,
    ),
    DocCfg(
        id='EFTA01361270',
        author=DEUTSCHE_BANK,
        date='2014-01-02',
        description=f"$60,000 transfer from {SOUTHERN_TRUST_COMPANY} to {BEN_GOERTZEL}'s Novamente",
    ),
    DocCfg(
        id='EFTA01111057',
        author=MORTIMER_ZUCKERMAN,
        date='2014-07-10',
        description='Mortimer B. Zuckerman Management Trust',
        non_participants=['Marla Maples'],
    ),
    DocCfg(id='EFTA01285411', description=f"bank statement for Epstein's {SOUTHERN_TRUST_COMPANY} showing $82 million balance"),
    DocCfg(id='EFTA01222951', description=f"credit card expenses for Carlos L Rodriguez using Plum Card", date='2019-02-12'),
    DocCfg(id='EFTA00016884', description="Epstein's last will and testament", date='2014-11-18'),
    DocCfg(
        id='EFTA00089546',
        description=f"Epstein last will and testament codicil naming {JAMES_CAYNE} an executor",
        date='2007-09-20',
        non_participants=[JOI_ITO],
    ),
    DocCfg(id='EFTA01266380', description="Epstein's 2014 Trust with bequests"),
    DocCfg(id='EFTA01282282', description=f"Epstein Butterfly Trust (sole beneficiary {KARYNA_SHULIAK})"),
    DocCfg(id='EFTA01583819', description=f"Epstein had control of {JAMES_CAYNE}'s assets"),
    DocCfg(id='EFTA00099424', description=f"Epstein 2017 Trust (Eva Andersson Dubin, {DARREN_INDYKE}, {RICHARD_KAHN})"),
    DocCfg(id='EFTA01266457', description=f"Epstein 2018 Trust ({KATHRYN_RUEMMLER}, {DARREN_INDYKE}, {RICHARD_KAHN})"),
    DocCfg(id='EFTA01266204', description=f"Epstein The 1953 Trust ({DARREN_INDYKE}, {RICHARD_KAHN})", date='2019-08-08'),
    DocCfg(id='EFTA00299927', description=f"estate plan for {JAMES_CAYNE} found in Epstein's possession"),
    DocCfg(id='EFTA01265973', description="large transfers around time of Epstein arrest", show_full_panel=True),
    DocCfg(id='EFTA01087311', description=f'{LEON_BLACK} Family Partners cash projections'),
    DocCfg(id='EFTA01086463', description=f"{MORTIMER_ZUCKERMAN}'s art collection valuations", is_valid_for_name_scan=False),
    DocCfg(id='EFTA00007781', description='paychecks signed by Epstein deposited at Colonial Bank', date='2005-08-12'),
    DocCfg(id='EFTA01273102', description=f"payment from Epstein to {RENATA_BOLOTOVA}'s father's account at Sberbank"),
    DocCfg(id='EFTA00238499', description='wire transfer to Signature Bank account'),
    DocCfg(id='EFTA00000476', replace_text_with='photo of JEFFREY EPSTEIN CASH DISBURSEMENTS for the month 2006-09', date='2006-09-01'),
    # Emails
    EmailCfg(id='EFTA00037187', is_interesting=True),
    EmailCfg(id='EFTA02189550', author='Ike Groff', author_uncertain=True, description=f"Ike Groff invests $250,000 in Mangrove Partners managed by Nathaniel August"),
    EmailCfg(id='EFTA00371120', description=f"Epstein appears to invest in {ATORUS}"),
    EmailCfg(id='EFTA00652799', description=f'Epstein calls Ari Glass "a bit sketchy" despite investing ~$50 million in his fund Boothbay'),
]



"""
Resumes and application letters.
"""
from epstein_files.documents.config.doc_cfg import DocCfg
from epstein_files.people.names import *
from epstein_files.util.constant.strings import *

HBS_APPLICATION = f"{HARVARD} Business School application letter"
RESUME_OF = 'professional resumé'


RESUMÉ_CFGS = [
    DocCfg(
        id='029304',
        attached_to_email_id='029299',
        author=MICHAEL_J_BOCCIO,
        description=f"recommendation letter by {DONALD_TRUMP}",
    ),
    DocCfg(id='022367', author='Jack J. Grynberg', description=RESUME_OF, date='2014-07-01'),
    DocCfg(
        id='029302',
        attached_to_email_id='029299',
        author=MICHAEL_J_BOCCIO,
        description=f"{RESUME_OF} (former lawyer at the {TRUMP_ORG})",
        date='2011-08-07',
    ),
    DocCfg(
        id='029301',
        attached_to_email_id='029299',
        author=MICHAEL_J_BOCCIO,
        description=f"letter from former lawyer at the {TRUMP_ORG}",
        date='2011-08-07',
    ),
    DocCfg(id='029102', author=NERIO_ALESSANDRI, description=HBS_APPLICATION),
    DocCfg(id='029104', author=NERIO_ALESSANDRI, description=HBS_APPLICATION),
    DocCfg(id='015671', author='Robin Solomon', description=RESUME_OF, date='2015-06-02'),  # She left Mount Sinai at some point in 2015,
    DocCfg(id='015672', author='Robin Solomon', description=RESUME_OF, date='2015-06-02'),  # She left Mount Sinai at some point in 2015,
    DocCfg(id='029623', description=f'bio of Kathleen Harrington, Founding Partner, C/H Global Strategies'),
]
